# -*- coding: utf-8 -*-
"""
Created on Wed Jun 12 18:12:43 2020

@author: Didou09
"""


import numpy as np
import datastock as ds


from ..geom._comp_solidangles import calc_solidangle_apertures


# ##################################################################
# ##################################################################
#             solid angles from any points
# ##################################################################


def compute_solid_angles(
    coll=None,
    key=None,
    key_cam=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # options
    config=None,
    visibility=None,
    # return
    return_vect=None,
    return_alpha=None,
):
    # ---------
    # check

    (
        key, key_cam, spectro,
        ptsx, ptsy, ptsz, shape0_pts,
        return_vect, return_alpha,
    ) = _compute_solid_angles_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        # pts
        ptsx=ptsx,
        ptsy=ptsy,
        ptsz=ptsz,
        # return
        return_vect=return_vect,
        return_alpha=return_alpha,
    )

    # -----------
    # prepare

    if spectro:
        raise NotImplementedError()

    else:

        dout = _compute_solid_angles_broadband(
            coll=coll,
            key=key,
            key_cam=key_cam,
            # pts
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            shape0_pts=shape0_pts,
            # options
            config=config,
            visibility=visibility,
            # return
            return_vect=return_vect,
        )

    return dout


# ##################################################################
#             check inputs
# ##################################################################


def _compute_solid_angles_check(
    coll=None,
    key=None,
    key_cam=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # options
    config=None,
    visibility=None,
    # return
    return_vect=None,
    return_alpha=None,
):
    # ---------
    # check

    # key_cam
    key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=key_cam)
    spectro = coll.dobj['diagnostic'][key]['spectro']

    # pts
    ptsx = np.atleast_1d(ptsx)
    ptsy = np.atleast_1d(ptsy)
    ptsz = np.atleast_1d(ptsz)

    if not (ptsx.shape == ptsy.shape == ptsz.shape):
        msg = (
            "Args ptsx, ptsy, ptsz must be 3 np.ndarray of the same shape!"
        )
        raise Exception(msg)

    shape0_pts = ptsx.shape
    if ptsx.ndim > 1:
        ptsx = ptsx.ravel()
        ptsy = ptsy.ravel()
        ptsz = ptsz.ravel()

    # return_vect
    return_vect = ds._generic_check._check_var(
        return_vect, 'return_vect',
        types=bool,
        default=False,
    )

    # return_alpha
    return_alpha = ds._generic_check._check_var(
        return_alpha, 'return_alpha',
        types=bool,
        default=False,
    )

    return (
        key, key_cam, spectro,
        ptsx, ptsy, ptsz, shape0_pts,
        return_vect, return_alpha,
    )


# ##################################################################
#             regular
# ##################################################################


def _compute_solid_angles_broadband(
    coll=None,
    key=None,
    key_cam=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    shape0_pts=None,
    # options
    config=None,
    visibility=None,
    # return
    return_vect=None,
):

    doptics = coll.dobj['diagnostic'][key]['doptics']
    dout = {k0: {} for k0 in key_cam}

    for k0 in key_cam:

        # prepare apertures
        dap = {}
        for op, opc in zip(doptics[k0]['optics'], doptics[k0]['cls']):
            dg = coll.dobj[opc][op]['dgeom']
            if dg['type'] == '3d':
                px, py, pz = dg['poly_x'], dg['poly_y'], dg['poly_z']
            else:
                cc = dg['cent']
                out0, out1 = dg['outline']
                out0, out1 = coll.ddata[out0]['data'], coll.ddata[out1]['data']
                px = cc[0] + out0*dg['e0'][0] + out1*dg['e1'][0]
                py = cc[1] + out0*dg['e0'][1] + out1*dg['e1'][1]
                pz = cc[2] + out0*dg['e0'][2] + out1*dg['e1'][2]

            dap[op] = {
                'nin': dg['nin'],
                'poly_x': px,
                'poly_y': py,
                'poly_z': pz,
            }

        # prepare camera
        dg = coll.dobj['camera'][k0]['dgeom']
        ddet = {}

        # cents
        cx, cy, cz = coll.get_camera_cents_xyz(k0)
        sh = cx.shape
        ddet['cents_x'] = cx
        ddet['cents_y'] = cy
        ddet['cents_z'] = cz

        # vectors
        ddet.update(coll.get_camera_unit_vectors(k0))
        for k1 in ['nin', 'e0', 'e1']:
            for ii, ss in enumerate(['x', 'y', 'z']):
                kk = f'{k1}_{ss}'
                if np.isscalar(ddet[kk]):
                    ddet[kk] = np.full(sh, ddet[kk])

        out0, out1 = dg['outline']
        out0, out1 = coll.ddata[out0]['data'], coll.ddata[out1]['data']
        ddet['outline_x0'] = out0
        ddet['outline_x1'] = out1

        # compute
        out = calc_solidangle_apertures(
            # observation points
            pts_x=ptsx,
            pts_y=ptsy,
            pts_z=ptsz,
            # polygons
            apertures=dap,
            detectors=ddet,
            # possible obstacles
            config=config,
            # parameters
            summed=False,
            visibility=visibility,
            return_vector=return_vect,
            return_flat_pts=True,
            return_flat_det=None,
        )

        # store
        if return_vect is True:
            dout[k0]['solid_angle'] = out[0]
            dout[k0]['vectx'] = out[1]
            dout[k0]['vecty'] = out[2]
            dout[k0]['vectz'] = out[3]
        else:
            dout[k0]['solid_angle'] = out

        # reshape
        if shape0_pts != ptsx.shape:
            shape = tuple(np.r_[dout[k0]['solid_angle'].shape[0], shape0_pts])
            for k1, v1 in dout[k0].items():
                dout[k0][k1] = v1.reshape(shape)

    return dout


"""Functions that are used in different tests"""
import numpy as np


def compute_ves_norm(ves_poly):
    ves_norm = ves_poly[:, 1:] - ves_poly[:, :-1]
    ves_norm = np.array([-ves_norm[1, :], ves_norm[0, :]])
    ves_norm = ves_norm / np.sqrt(np.sum(ves_norm ** 2, axis=0))[np.newaxis, :]
    return ves_norm


def compute_min_max_r_and_z(ves_poly):
    r_min_max = np.array([np.min(ves_poly[0, :]), np.max(ves_poly[0, :])])
    z_min_max = np.array([np.min(ves_poly[1, :]), np.max(ves_poly[1, :])])
    return r_min_max, z_min_max



"""
Computing a camera image with custom emissivity
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This tutorial defines an emissivity that varies in space and computes the
signal received by a camera using this emissivity.
"""

###############################################################################
# We start by loading a built-in `tofu` configuration and define a 2D camera.

import matplotlib.pyplot as plt
import numpy as np
import tofu as tf

configB2 = tf.load_config("B2")

cam2d = tf.geom.utils.create_CamLOS2D(
    config=configB2,
    pinhole=[3.4, 0, 0],
    sensor_nb=100,
    focal=0.1,
    sensor_size=0.1,
    orientation=[np.pi, np.pi/6, 0],
    Name="",
    Exp="",
    Diag="",
)

###############################################################################
# Now, we define an emissivity function that depends on r and z coordinates.
# We can plot its profile in the (0, X, Z) plane.


def emissivity(pts, t=None, vect=None):
    """Custom emissivity as a function of geometry.

    :param pts: ndarray of shape (3, npts) (each column is a xyz coordinate)
    :param t: optional, time parameter to add a time dependency to the
        emissivity function
    :param vect: optional, ndarray of shape (3, npts), if anisotropic
        emissivity, unit direction vectors (X,Y,Z)
    :return:
        - emissivity -- 2D array holding the emissivity for each point in the
            input grid
    """
    r, z = np.hypot(pts[0, :], pts[1, :]), pts[2, :]
    e = np.exp(-(r - 2.4) ** 2 / 0.2 ** 2 - z ** 2 / 0.4 ** 2)
    if t is not None:
        e = np.cos(np.atleast_1d(t))[:, None] * e[None, :]
    else:
        # as stated in documentation of calc_signal, e.ndim must be 2
        e = np.reshape(e, (1, -1))
    return e


y = np.linspace(2, 3, num=90)
z = np.linspace(-0.5, 0.5, num=100)
Y, Z = np.meshgrid(y, z)
X = np.zeros_like(Y)
pts = np.c_[X.ravel(), Y.ravel(), Z.ravel()].T
emissivity_vals = emissivity(pts)
emissivity_vals = emissivity_vals.reshape(X.shape)


def project_to_2D(xyz):
    """Projection to (0, X, Z) plane."""
    return xyz[0], xyz[2]


fig, ax = plt.subplots()
ax.pcolormesh(Y, Z, emissivity_vals)
ax.set_xlabel('y')
ax.set_ylabel('z')
configB2.plot(lax=ax, proj='cross')
cam_center, = ax.plot(*project_to_2D(cam2d._dgeom['pinhole']), '*', ms=20)
ax.set_aspect("equal")
ax.legend(handles=[cam_center], labels=['camera pinhole'], loc='upper right')

###############################################################################
# Finally, we compute an image using the 2D camera and this emissivity.
# If we provide a time vector, the field will vary in a cosinusoidal fashion
# (see above definition) across time.

time_vector = np.linspace(0, 2 * np.pi, num=100)

sig, units = cam2d.calc_signal(emissivity,
                               res=0.01,
                               reflections=False,
                               minimize="hybrid",
                               method="sum",
                               newcalc=True,
                               plot=False,
                               ani=False,
                               t=time_vector)

sig.plot(ntMax=1)
plt.show(block=False)



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds


# specific
from . import _generic_check
from . import _generic_plot
from . import _class8_plot as _plot


# ###############################################################
# ###############################################################
#                           plot main
# ###############################################################


def _plot(
    coll=None,
    key=None,
    key_cam=None,
    doptics=None,
    is2d=None,
    spectro=None,
    elements=None,
    proj=None,
    los_res=None,
    indch=None,
    indlamb=None,
    # data plot
    dvos=None,
    dplot=None,
    units=None,
    cmap=None,
    dvminmax=None,
    alpha=None,
    # camera
    x0=None,
    x1=None,
    out0=None,
    out1=None,
    extent_cam=None,
    # los, vos
    los_x=None,
    los_y=None,
    los_z=None,
    los_r=None,
    los_xi=None,
    los_yi=None,
    los_zi=None,
    los_ri=None,
    pc0=None,
    pc1=None,
    ph0=None,
    ph1=None,
    pc0i=None,
    pc1i=None,
    ph0i=None,
    ph1i=None,
    # mesh
    p0=None,
    p1=None,
    etendue=None,
    length=None,
    # config
    plot_config=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    # interactivity
    color_dict=None,
    **kwdargs,
):

    # ---------------------
    # prepare ph_counts, lamb, cos...

    ddata, extent, dhor = _prepare_ph(
        coll=coll,
        dvos=dvos[key_cam[0]],
        key=key,
        key_cam=key_cam,
        indch=indch,
        indlamb=indlamb,
        spectro=spectro,
        is2d=is2d,
        etendue=etendue,
        length=length,
    )

    # vmin, vmax
    dvminmax = _get_dvminmax(
        ddata=ddata,
        dvminmax=dvminmax,
    )

    # -----------------
    # prepare figure

    if dax is None:

        dax = _get_dax(
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            tit=key,
            is2d=is2d,
            key_cam=key_cam,
            indch=indch,
            ddata=ddata,
        )

    dax = _generic_check._check_dax(dax=dax, main=proj[0])

    # -----------------
    # plot diag elements

    for k0, v0 in dplot.items():

        for k1, v1 in v0.items():

            # cross
            typ = 'cross'
            lkax = [kk for kk, vv in dax.items() if vv['type'] == typ]
            for kax in lkax:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['r'],
                        v1['z'],
                        v1['ur'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['r'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # hor
            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['ux'],
                        v1['uy'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        **v1.get('props', {}),
                    )

    # ------------------
    # plot ddata

    for k0, v0 in ddata.items():

        kax = [kk for kk, vv in dax.items() if vv['k0'] == k0][0]
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            typ = dax[kax]['type']
            if typ == 'camera':
                im = ax.imshow(
                    v0['data'].T,
                    extent=extent_cam,
                    origin='lower',
                    aspect='equal',
                    interpolation='nearest',
                    vmin=dvminmax[k0][0],
                    vmax=dvminmax[k0][1],
                )

                # camera outline
                ax.plot(
                    np.r_[out0, out0[0]],
                    np.r_[out1, out1[0]],
                    c='k',
                    ls='-',
                    lw=1,
                )

                # pixel
                ax.plot(
                    [x0[indch[0]]],
                    [x1[indch[1]]],
                    c='k',
                    ls='None',
                    marker='s',
                    ms=6,
                )

            elif typ == 'cross':
                im = ax.imshow(
                    v0['data'].T,
                    extent=extent,
                    origin='lower',
                    aspect='equal',
                    interpolation='nearest',
                    vmin=dvminmax[k0][0],
                    vmax=dvminmax[k0][1],
                )

                # ------------------
                # plot los / vos

                _add_camera_los_cross(
                    ax=ax,
                    is2d=is2d,
                    los_r=los_r,
                    los_z=los_z,
                    pc0=pc0,
                    pc1=pc1,
                    alpha=alpha,
                    color_dict=color_dict,
                )

                _add_camera_los_cross(
                    ax=ax,
                    is2d=is2d,
                    los_r=los_ri,
                    los_z=los_zi,
                    pc0=pc0i,
                    pc1=pc1i,
                    alpha=alpha,
                    color_dict=color_dict,
                )

                # mesh outline
                if dax[kax]['jj'] == 1:
                    ax.plot(
                        np.r_[p0, p0[0]],
                        np.r_[p1, p1[0]],
                        c='k',
                        ls='-',
                        lw=1,
                    )

            # labels
            if dax[kax]['ii'] > 0:
                ax.tick_params(labelleft=False)
            if dax[kax]['jj'] == 1:
                ax.tick_params(labelbottom=False)

            plt.colorbar(im, ax=ax)

    # hor
    kax = 'hor'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _add_camera_los_hor(
            ax=ax,
            is2d=is2d,
            los_x=los_xi,
            los_y=los_yi,
            ph0=ph0,
            ph1=ph1,
            alpha=alpha,
            color_dict=color_dict,
        )

        ax.plot(
            dhor['envelopx'],
            dhor['envelopy'],
            ls='-',
            c='k',
            marker='.',
            ms=4,
            lw=1.
        )

        ax.fill(
            dhor['minx'],
            dhor['miny'],
            ls='None',
            lw=1.,
            fc='r',
            alpha=0.5,
        )

        ax.plot(
            dhor['mainx'],
            dhor['mainy'],
            ls='-',
            c='r',
            marker='.',
            ms=4,
            lw=1.
        )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        typ = 'cross'
        lkax = [kk for kk, vv in dax.items() if vv['type'] == typ]
        for kax in lkax:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj='cross', dLeg=False)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

    # -------
    # connect

    return dax


# ##################################################################
# ##################################################################
#                       Prepare sang
# ##################################################################


def _prepare_ph(
    coll=None,
    dvos=None,
    key=None,
    key_cam=None,
    indch=None,
    indlamb=None,
    spectro=None,
    is2d=None,
    etendue=None,
    length=None,
):

    # -----------------
    # check up

    if indlamb is None:
        indlamb = int(dvos['lamb']['data'].size / 2)

    # -----------------
    # get mesh sampling

    dsamp = coll.get_sample_mesh(
        key=dvos['keym'],
        res=dvos['res_RZ'],
        mode='abs',
        grid=False,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    # ------------------------------
    # dhor

    x0u = dsamp['x0']['data']
    x1u = dsamp['x1']['data']

    iru = np.unique(dvos['indr_cross']['data'])
    phi_env = np.full((iru.size, 2), np.nan)
    phi_mean = np.full((iru.size,), np.nan)
    phi_minmax = np.full((iru.size, 2), np.nan)
    for ii, i0 in enumerate(iru):
        ind = dvos['indr_cross']['data'] == i0
        phi_env[ii, 0] = np.nanmin(dvos['phi_min']['data'][..., ind])
        phi_env[ii, 1] = np.nanmax(dvos['phi_max']['data'][..., ind])

        phi_mean[ii] = np.nanmean(dvos['phi_mean']['data'][indch[0], indch[1], ind])
        phi_minmax[ii, 0] = np.nanmin(dvos['phi_min']['data'][indch[0], indch[1], ind])
        phi_minmax[ii, 1] = np.nanmax(dvos['phi_max']['data'][indch[0], indch[1], ind])

    dhor = {
        'envelopx': np.r_[
            x0u[iru] * np.cos(phi_env[:, 0]),
            x0u[iru[::-1]] * np.cos(phi_env[::-1, 1]),
            x0u[iru[0]] * np.cos(phi_env[0, 0]),
        ],
        'envelopy': np.r_[
            x0u[iru] * np.sin(phi_env[:, 0]),
            x0u[iru[::-1]] * np.sin(phi_env[::-1, 1]),
            x0u[iru[0]] * np.sin(phi_env[0, 0]),
        ],
        'mainx': x0u[iru] * np.cos(phi_mean),
        'mainy': x0u[iru] * np.sin(phi_mean),
        'minx': np.r_[
            x0u[iru] * np.cos(phi_minmax[:, 0]),
            x0u[iru[::-1]] * np.cos(phi_minmax[::-1, 1]),
            x0u[iru[0]] * np.cos(phi_minmax[0, 0]),
        ],
        'miny': np.r_[
            x0u[iru] * np.sin(phi_minmax[:, 0]),
            x0u[iru[::-1]] * np.sin(phi_minmax[::-1, 1]),
            x0u[iru[0]] * np.sin(phi_minmax[0, 0]),
        ],
    }

    # ------------------------------
    # prepare image in cross-section

    n0, n1 = x0u.size, x1u.size
    shape = (n0, n1)

    nc_tot = np.full(shape, np.nan)
    nc_toti = np.full(shape, np.nan)

    dV = np.full(shape, np.nan)

    ph_tot = np.full(shape, np.nan)
    ph_toti = np.full(shape, np.nan)
    ph_tot_lamb = np.full(shape, np.nan)
    ph_toti_lamb = np.full(shape, np.nan)

    lamb = np.full(shape, np.nan)
    lambi = np.full(shape, np.nan)
    dlamb = np.full(shape, np.nan)
    dlambi = np.full(shape, np.nan)
    cos = np.full(shape, np.nan)
    cosi = np.full(shape, np.nan)

    shape_cam = dvos['ph']['data'].shape[:-2]
    nc_cam = np.full(shape_cam, np.nan)
    ph_cam = np.full(shape_cam, np.nan)
    cos_cam = np.full(shape_cam, np.nan)
    lamb_cam = np.full(shape_cam, np.nan)
    dlamb_cam = np.full(shape_cam, np.nan)

    indr = dvos['indr_cross']['data']
    indz = dvos['indz_cross']['data']
    npts, nlamb = dvos['ph']['data'].shape[-2:]

    # multiply by dlamb
    ph0 = dvos['ph']['data'] * np.mean(np.diff(dvos['lamb']['data']))
    if is2d:
        nc = dvos['ncounts']['data'].reshape((-1, npts))
        ph = ph0.reshape((-1, npts, nlamb))
        coss = dvos['cos']['data'].reshape((-1, npts))
        lambc = dvos['lamb']['data'][None, None, None, :]
        indch = indch[0] * shape_cam[1] + indch[1]

    else:
        nc = dvos['ncounts']['data']
        ph = ph0
        coss = dvos['cos']['data']
        lambc = dvos['lamb']['data'][None, None, :]

    nci = nc[indch, :]
    phi = ph[indch, :, :]
    cossi = coss[indch, :]

    # lambmax, lambmin
    lambf = np.repeat(dvos['lamb']['data'][None, :], npts, axis=0)
    for ni in shape_cam[::-1]:
        lambf = np.repeat(lambf[None, :], ni, axis=0)

    ind0 = ph0 == 0
    lambf[ind0] = np.inf
    lambmin = np.min(lambf, axis=-1)
    lambf[ind0] = -np.inf
    lambmax = np.max(lambf, axis=-1)
    del lambf

    lambmini = lambmin.reshape((-1, npts))[indch]
    lambmaxi = lambmax.reshape((-1, npts))[indch]

    # photon counts
    nc_tot[indr, indz] = np.sum(nc, axis=0)
    nc_toti[indr, indz] = np.sum(nci, axis=0)

    # update indr, indz
    iok = nc_tot[indr, indz] > 0.
    ir, iz = indr[iok], indz[iok]
    ioki = nc_toti[indr, indz] > 0.
    iri, izi = indr[ioki], indz[ioki]

    nc_tot[nc_tot == 0] = np.nan
    nc_toti[nc_toti == 0] = np.nan

    # dV
    dV[ir, iz] =  dvos['dV']['data'][iok]

    # cos
    cos[ir, iz] = np.nansum(nc[:, iok] * coss[:, iok], axis=0) / nc_tot[ir, iz]
    cosi[iri, izi] = cossi[ioki]

    # ph_tot
    if np.any(np.isnan(ph)):
        msg = "ph_count should not contain nans! (nansum copies)"
        raise Exception(msg)

    ph_tot[ir, iz] = np.sum(np.sum(ph[:, iok, :], axis=0), axis=-1)
    ph_toti[iri, izi] = np.sum(phi[ioki, :], axis=-1)

    # average wavelength
    lamb[ir, iz] = np.sum(
        np.sum(ph[:, iok, :] * dvos['lamb']['data'][None, None, :], axis=0),
        axis=-1,
    ) / ph_tot[ir, iz]

    lambi[iri, izi] = (
        np.sum(phi[ioki, :] * dvos['lamb']['data'][None, :], axis=-1)
        / ph_toti[iri, izi]
    )

    # delta wavelength
    dlamb[ir, iz] = (
        np.nanmax(lambmax.reshape((-1, npts))[:, iok], axis=0)
        - np.nanmin(lambmin.reshape((-1, npts))[:, iok], axis=0)
    )
    dlambi[iri, izi] = lambmaxi[ioki] - lambmini[ioki]

    # adjust
    ph_tot[ph_tot == 0.] = np.nan
    ph_toti[ph_toti == 0.] = np.nan

    # -----------------------
    # prepare image on camera

    nc_cam[...] = np.nansum(dvos['ncounts']['data'], axis=-1)
    iok = nc_cam > 0.
    nc_cam[~iok] = np.nan

    ph_cam[iok] = np.nansum(np.nansum(ph0, axis=-1), axis=-1)[iok]

    lamb_cam[iok] = (
        np.nansum(np.nansum(ph0 * lambc, axis=-1), axis=-1)[iok]
        / ph_cam[iok]
    )

    cos_cam[iok] = (
        np.nansum(dvos['ncounts']['data'] * dvos['cos']['data'], axis=-1)[iok]
        / nc_cam[iok]
    )

    iok = np.any(np.isfinite(lambmax), axis=-1)
    dlamb_cam[iok] = (
        np.nanmax(lambmax[iok, :], axis=-1)
        - np.nanmin(lambmin[iok, :], axis=-1)
    )

    # ----------------------
    # prepare per wavelength

    ph_cam_lamb = np.nansum(ph0[..., indlamb], axis=-1)
    ph_tot_lamb[indr, indz] = np.nansum(ph[..., indlamb], axis=0)
    ph_toti_lamb[indr, indz] = np.nansum(phi[..., indlamb], axis=0)

    ph_cam_lamb[ph_cam_lamb == 0] = np.nan

    # ----------------------
    # delta_lamb for etendue

    delta_lamb = coll.get_diagnostic_data(key, data='dlamb')[0][key_cam[0]]

    # -------------------
    # extent

    x0 = dsamp['x0']['data']
    dx0 = x0[1] - x0[0]
    x1 = dsamp['x1']['data']
    dx1 = x1[1] - x1[0]

    extent = (
        x0[0] - 0.5*dx0,
        x0[-1] + 0.5*dx0,
        x1[0] - 0.5*dx1,
        x1[-1] + 0.5*dx1,
    )

    return (
        {
            'dV': {'data': dV},
            'nc_tot': {'data': nc_tot},
            'nc_toti': {'data': nc_toti},
            'ph_tot': {'data': ph_tot},
            'ph_toti': {'data': ph_toti},
            'cos': {'data': cos},
            'lamb': {'data': lamb},
            'dlamb': {'data': dlamb},
            'cosi': {'data': cosi},
            'lambi': {'data': lambi},
            'dlambi': {'data': dlambi},
            'nc_cam': {'data': nc_cam},
            'ph_cam': {'data': ph_cam},
            'cos_cam': {'data': cos_cam},
            'lamb_cam': {'data': lamb_cam},
            'dlamb_cam': {'data': dlamb_cam},
            'ph_cam_lamb': {'data': ph_cam_lamb},
            'ph_tot_lamb': {'data': ph_tot_lamb},
            'ph_toti_lamb': {'data': ph_toti_lamb},
            'etendue*length*dlamb': {'data': etendue * length * delta_lamb},
            # 'etendue*length': {'data': etendue * length},
        },
        extent,
        dhor,
    )


# ###############################################################
# ###############################################################
#                       vminmax
# ###############################################################


def _get_dvminmax(ddata=None, dvminmax=None):

    if dvminmax is None:
        dvminmax = {}

    # safety check
    lkout = [k0 for k0 in dvminmax.keys() if k0 not in ddata.keys()]
    if len(lkout) > 0:
        lstr_in = [f"\t{k0}" for k0 in ddata.keys()]
        lstr_out = [f"\t{k0}" for k0 in lkout]
        msg = (
            "Arg dvminmax must be a dict with the following keys only:\n"
            + '\n'.join(lstr_in)
            + "\nHence, the following ones are not valid:\n"
            + '\n'.join(lstr_out)
        )
        raise Exception(msg)

    # fill with default
    for k0, v0 in ddata.items():
        if dvminmax.get(k0) is None:
            dvminmax[k0] = [None, None]

        if dvminmax[k0][0] is None:
            dvminmax[k0][0] = np.nanmin(v0['data'])
        if dvminmax[k0][1] is None:
            dvminmax[k0][1] = np.nanmax(v0['data'])

    return dvminmax


# ##################################################################
# ##################################################################
#                       add mobile
# ##################################################################


def _add_camera_los_cross(
    ax=None,
    is2d=None,
    color_dict=None,
    los_r=None,
    los_z=None,
    pc0=None,
    pc1=None,
    alpha=None,
):

    # ------
    # los

    if los_r is not None:
        l0, = ax.plot(
            los_r,
            los_z,
            c='k',
            ls='-',
            lw=1.,
        )

    # ------
    # vos
    if pc0 is not None:
        if pc0.ndim == 2:
            for ii in range(pc0.shape[1]):
                l0, = ax.fill(
                    pc0[:, ii],
                    pc1[:, ii],
                    fc='k',
                    alpha=alpha,
                    ls='None',
                    lw=0.,
                )

        elif pc0.ndim == 3:
            for ii in range(pc0.shape[1]):
                for jj in range(pc0.shape[2]):
                    l0, = ax.fill(
                        pc0[:, ii, jj],
                        pc1[:, ii, jj],
                        fc='k',
                        alpha=alpha,
                        ls='None',
                        lw=0.,
                    )

        else:
            l0, = ax.fill(
                pc0,
                pc1,
                fc='k',
                alpha=alpha,
                ls='None',
                lw=0.,
            )


def _add_camera_los_hor(
    ax=None,
    is2d=None,
    los_x=None,
    los_y=None,
    ph0=None,
    ph1=None,
    alpha=None,
    color_dict=None,
):

    # ------
    # los

    if los_x is not None:
        l0, = ax.plot(
            los_x,
            los_y,
            c=color_dict['x'][0],
            ls='-',
            lw=1.,
        )

    # ------
    # vos

    if ph0 is not None:

        if ph0.ndim == 2:
            for ii in range(ph0.shape[1]):
                l0, = ax.fill(
                    ph0[:, ii],
                    ph1[:, ii],
                    fc='k',
                    alpha=alpha,
                    ls='None',
                    lw=0.,
                )

        elif ph0.ndim == 3:
            for ii in range(ph0.shape[1]):
                for jj in range(ph0.shape[2]):
                    l0, = ax.fill(
                        ph0[:, ii, jj],
                        ph1[:, ii, jj],
                        fc='k',
                        alpha=alpha,
                        ls='None',
                        lw=0.,
                    )

        else:
            l0, = ax.fill(
                ph0,
                ph1,
                fc='k',
                alpha=alpha,
                ls='None',
                lw=0.,
            )


def _add_camera_data(
    coll=None,
    ax=None,
    sang_integ=None,
    etendue=None,
    length=None,
    indch=None,
    color_dict=None,
    # vmin, vmax
    vmin_cam=None,
    vmax_cam=None,
    # 2d only
    ax_diff=None,
    ax_etend=None,
    is2d=None,
    x0=None,
    x1=None,
    extent_cam=None,
):


    if is2d:

        # sang
        mi = ax.imshow(
            sang_integ.T,
            origin='lower',
            extent=extent_cam,
            interpolation='nearest',
            vmin=vmin_cam,
            vmax=vmax_cam,
        )

        # etendue
        etendle = etendue * length
        mi = ax_etend.imshow(
            etendle.T,
            origin='lower',
            extent=extent_cam,
            interpolation='nearest',
            vmin=vmin_cam,
            vmax=vmax_cam,
        )
        plt.colorbar(mi, ax=[ax, ax_etend])

        # diff
        diff = (sang_integ - etendle).T
        dmax = np.abs(max(np.nanmin(diff), np.nanmax(diff)))
        imd = ax_diff.imshow(
            (sang_integ - etendle).T,
            origin='lower',
            extent=extent_cam,
            interpolation='nearest',
            cmap=plt.cm.seismic,
            vmin=-dmax,
            vmax=dmax,
        )

        plt.colorbar(imd, ax=ax_diff)

        # marker
        for aa in [ax, ax_etend, ax_diff]:
            aa.plot(
                [x0[indch[0]]],
                [x1[indch[1]]],
                c='k',
                marker='s',
                ms=6,
                ls='None',
                lw=1.,
            )

    else:

        nlos = sang_integ.shape[0]
        ind = np.arange(0, nlos)

        # vos
        ax.plot(
            ind,
            sang_integ,
            c='b',
            marker='.',
            ls='-',
            lw=1.,
            label="sang * dV (sr.m3)",
        )

        # los
        ax.plot(
            ind,
            etendue * length,
            c='k',
            marker='.',
            ls='-',
            lw=1.,
            label="etendue * length (sr.m3)",
        )

        # indch
        ax.axvline(
            indch,
            c='k',
            lw=1.,
            ls='--',
        )

        ax.set_ylim(vmin_cam, vmax_cam)
        plt.legend()


# ################################################################
# ################################################################
#                   figure
# ################################################################


def _get_dax(
    proj=None,
    dmargin=None,
    fs=None,
    tit=None,
    wintit=None,
    is2d=None,
    key_cam=None,
    indch=None,
    ddata=None,
):

    # ------------
    # check inputs

    # fs
    if fs is None:
        fs = (16, 9)

    # dmargin
    dmargin = ds._generic_check._check_var(
        dmargin, 'dmargin',
        types=dict,
        default={
            'bottom': 0.05, 'top': 0.90,
            'left': 0.04, 'right': 0.98,
            'wspace': 0.20, 'hspace': 0.40,
            # 'width_ratios': [0.6, 0.4],
            # 'height_ratios': [0.4, 0.6],
        },
    )

    # wintit
    wintit = ds._generic_check._check_var(
        wintit, 'wintit',
        types=str,
        default=_generic_plot._WINDEF,
    )

    # tit
    if tit is None:
        tit = f"{key_cam} - {indch}\nVOS"

    # -------------
    # Prepare

    lcross = [
        'dV',
        'ph_tot', 'ph_toti',
        'nc_tot', 'nc_toti',
        'cos', 'cosi',
        'lamb', 'lambi',
        'dlamb', 'dlambi',
        'ph_tot_lamb', 'ph_toti_lamb',
    ]
    lcross = [k0 for k0 in lcross if k0 in ddata.keys()]
    lcross.insert(1, None)

    lcam = [
        'ph_cam', 'nc_cam', 'cos_cam', 'lamb_cam', 'dlamb_cam',
        'ph_cam_lamb',
    ]
    lcam = [k0 for k0 in lcam if k0 in ddata.keys()]

    ncross = len(lcross)
    ncam = len(lcam)

    # -------------
    # Create figure

    fig = plt.figure(figsize=fs)
    fig.canvas.manager.set_window_title(wintit)
    fig.suptitle(tit, size=12, fontweight='bold')

    if is2d is False:
        raise NotImplementedError()

    gs = gridspec.GridSpec(ncols=ncam + 1, nrows=3, **dmargin)

    dax = {}
    shxca, shyca = None, None
    shxcr, shycr = None, None
    for ii in range(0, ncam + 1):
        for jj in range(3):

            # prepare share options
            if ii == 0 and jj == 2:
                shx = None
                shy = None
            elif jj == 0:
                shx = shxca
                shy = shyca
            else:
                shx = shxcr
                shy = shycr

            # create axes
            ax = fig.add_subplot(
                gs[jj, ii],
                aspect='equal',
                sharex=shx,
                sharey=shy,
            )

            # update share options
            if jj == 0 and shxca is None:
                shxca = ax
                shyca = ax

            if jj > 0 and shxcr is None:
                shxcr = ax
                shycr = ax

            # labels
            if jj == 0:
                if ii == 0:
                    name = 'etendue*length*dlamb'
                else:
                    name = lcam[ii -1].split('_')[0]
                    if name == 'ph':
                        name = 'photon flux'
                    if '_lamb' in lcam[ii - 1]:
                        name = f'{name}\n' + r'$\lambda$ = '

                ax.set_title(name, size=12, fontweight='bold')
                ax.set_xlabel('x0 (m)', size=12)
                if ii == 0:
                    ax.set_ylabel('x1 (m)', size=12)

            elif jj == 2:
                if ii == 0:
                    ax.set_xlabel('X (m)')
                    ax.set_ylabel('Y (m)')
                else:
                    ax.set_xlabel('R (m)')
                    if ii == 1:
                        ax.set_ylabel('Z (m)')
            elif jj == 1 and ii == 0:
                name = 'dV'
                ax.set_title(name, size=12, fontweight='bold')
                ax.set_ylabel('Z (m)')

            # update dict
            if ii == 0:
                if jj == 0:
                    kax = f'{name}_cam'
                    typ = 'camera'
                    k0 = name
                elif jj == 1:
                    kax = f'{name}_cross'
                    typ = 'cross'
                    k0 = 'dV'
                elif jj == 2:
                    kax = 'hor'
                    typ = 'hor'
                    k0 = None
            elif jj == 0:
                kax = f'{name}_cam'
                typ = 'camera'
                k0 = lcam[ii - 1]
            else:
                kax = f'{name}_cross'
                typ = 'cross'
                if jj == 2:
                    kax = f'{kax}_pix'

                k0 = lcross[ii*2 + (jj - 1)]

            dax[kax] = {
                'handle': ax, 'type': typ, 'k0': k0, 'ii': ii, 'jj': jj,
            }

    return dax


#!/usr/bin/env python

import os
import argparse

import numpy as np


_save = True
_here = os.path.abspath(os.path.dirname(__file__))
_Exp, _Cls, _name = os.path.split(__file__)[1].split('_')[:3]
assert not any([any([ss in s for ss in ['Notes','.']])
               for s in [_Exp, _Cls, _name]])

def get_notes():

    notes = {'DPhi':{}, 'dPhi':{}}
    # Total length
    notes['DL'] = 449.000

    # Toroidal width
    notes['DPhi']['In'] = 26.151
    notes['DPhi']['Out'] = 31.887

    # Toroidal gap
    notes['dPhi']['In'] = 0.681
    notes['dPhi']['Out'] = 0.680

    # Height
    notes['DZ'] = 41.562

    # Number in toroidal direction
    notes['nbPhi'] = 12*38

    # sample (X,Z,Y)
    notes['sampleXZY'] = [[-1917.967, 640.060, -293.245],
                          [-1932.817, 601.230, -299.739],
                          [-2167.242, 697.010, -335.545],
                          [-2168.645, 697.479, -335.689],
                          [-2198.613, 709.727, -340.266],
                          [-2201.495, 711.408, -340.635],
                          [-2344.246, 769.769, -362.438],
                          [-2329.510, 808.258, -356.103]]
    notes['sampleXZY'] = np.asarray(notes['sampleXZY'],dtype=float)

    for kk in notes.keys():
        if type(notes[kk]) is dict:
            notes[kk]['In'] = notes[kk]['In']*1.e-3
            notes[kk]['Out'] = notes[kk]['Out']*1.e-3
        elif not 'nb' in kk:
            notes[kk] = notes[kk]*1.e-3
    return notes


def make_Poly(save=_save, path=_here):

    notes = get_notes()

    Poly = np.array([np.hypot(notes['sampleXZY'][:,0],notes['sampleXZY'][:,2]),
                     notes['sampleXZY'][:,1]])
    Poly0 = Poly[:,[0,1,-2,-1]]
    PM = np.mean(Poly[:,[2,5]],axis=1)[:,np.newaxis]
    Poly1 = np.concatenate((Poly0[:,:2], PM,Poly0[:,2:]),axis=1)

    if save:
        cstr = '%s_%s_%s'%(_Exp,_Cls,_name)
        pathfilext = os.path.join(path, cstr+'_V0.txt')
        np.savetxt(pathfilext, Poly0.T)
        pathfilext = os.path.join(path, cstr+'_V1.txt')
        np.savetxt(pathfilext, Poly1.T)
        pathfilext = os.path.join(path, cstr+'_V2.txt')
        np.savetxt(pathfilext, Poly.T)
    return Poly0, Poly1, Poly, notes



if __name__=='__main__':

    # Parse input arguments
    msg = 'Launch creation of polygons txt from bash'
    parser = argparse.ArgumentParser(description = msg)

    parser.add_argument('-save', type=bool, help='save ?', default=_save)
    parser.add_argument('-path', type=str, help='saving path ?', default=_here)

    args = parser.parse_args()

    # Call wrapper function
    make_Poly(save=args.save, path=args.path)



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 15:36:19 2023

@author: dvezinet
"""


import os


import numpy as np
import datastock as ds


# ###############################################################
# ###############################################################
#                           Main
# ###############################################################


def add_camera(
    coll=None,
    cam=None,
    key=None,
):

    # -------------------
    # check inputs
    # -------------------

    coll, cam, key = _check(
        coll=coll,
        cam=cam,
        key=key,
    )

    # --------------------
    # extract geometry
    # --------------------

    dgeom, typ = _extract_dgeom(cam)

    # --------------------
    # extract material
    # --------------------

    dmat = _extract_dmat(cam)

    # --------------------
    # add
    # --------------------

    if typ == '1d':
        coll.add_camera_1d(
            key=key,
            dgeom=dgeom,
            dmat=dmat,
            color=None,
        )

    elif typ == '2d':
        coll.add_camera_2d(
            key=key,
            dgeom=dgeom,
            dmat=dmat,
            color=None,
        )

    else:
        raise NotImplementedError()

    return


# ###############################################################
# ###############################################################
#                           Check
# ###############################################################

def _check(
    coll=None,
    cam=None,
    key=None,
):


    # -------------------
    # check coll
    # -------------------

    import tofu as tf
    if coll is None:
        coll = tf.data.Collection()

    if not isinstance(coll, tf.data.Collection):
        msg = (
            "Arg coll must be a tf.data.Collection instance!\n"
            "Provided:\n"
            "\t- class: {type(coll)}\n"
            "\t- value: {coll}\n"
        )
        raise Exception(msg)

    # -------------------
    # check camera
    # -------------------

    # load from file
    if isinstance(cam, str):
        error = False
        if not (os.path.isfile(cam) and cam.endswith('.npz')):
            error = 'not a valid file name'
        else:
            try:
                cam = tf.load(cam)
            except Exception as err:
                try:
                    cam = dict(np.load(cam, allow_pickle=True))
                except Exception as err:
                    error = f'tf.load() failed to load {cam}\n{str(err)}'

        # raise error
        if error is not False:
            msg = (
                f"Arg cam must be a valid .npz file, loadable!\n{error}"
            )
            raise Exception(msg)

    # check class
    if isinstance(cam, dict):
        pass
    elif 'Camera' not in cam.__class__.__name__:
        msg = (
            "Arg cam must be a legacy tf.geom.CrystalBragg instance!\n"
            "Provided:\n"
            f"\t- class: {type(cam)}\n"
            f"\t- value: {cam}\n"
        )
        raise Exception(msg)

    # -------------------
    # check key
    # -------------------

    # default value
    if isinstance(cam, dict):
        kdef = "cam"
    else:
        kdef = cam.Id.Name

    # check
    lout = list(coll.dobj.get('camera', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        default=kdef,
        excluded=lout,
    )

    return coll, cam, key


# ###############################################################
# ###############################################################
#                       Extract dgeom
# ###############################################################


def _extract_dgeom(cam):


    if isinstance(cam, dict):

        typ = cam.get('type', '2d')

        x0 = cam['xi']
        x1 = cam['xj']

        dgeom = {
            'cent': np.copy(cam['cent']),
            'cents_x0': np.copy(x0),
            'cents_x1': np.copy(x1),
            'nin': np.copy(cam['nout']),
            'e0': np.copy(cam['ei']),
            'e1': np.copy(cam['ej']),
        }

        # outline of individual pixels
        if cam.get('outline_x0') is None:
            dx0 = 0.5 * np.mean(np.diff(x0))
            out0 = dx0 * np.r_[-1, 1, 1, -1]
            dx1 = 0.5 * np.mean(np.diff(x1))
            out1 = dx1 * np.r_[-1, -1, 1, 1]
            dgeom['outline_x0'] = out0
            dgeom['outline_x1'] = out1
        else:
            dgeom['outline_x0'] = np.copy(cam['outline_x0'])
            dgeom['outline_x1'] = np.copy(cam['outline_x1'])

    else:
        raise NotImplementedError()

    return dgeom, typ


# ###############################################################
# ###############################################################
#                       Extract dmat
# ###############################################################


def _extract_dmat(cam):

    # ------------
    # dmat

    dmat = {}

    return dmat


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 14 08:12:28 2024

@author: dvezinet
"""


# Built-in
import itertools as itt


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.gridspec as gridspec
import datastock as ds


from . import _class8_vos_utilities as _vos_utils


# ################################################################
# ################################################################
#                           Main
# ################################################################


def main(
    coll=None,
    key=None,
    # observation directions
    observation_directions=None,
    # mesh sampling
    key_mesh=None,
    res_RZ=None,
    nan0=None,
    # plotting options
    config=None,
    dcolor=None,
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
    cmap=None,
    vmin=None,
    vmax=None,
):

    # -------------
    # check inputs
    # -------------

    key, is_vos, observation_directions, keym, res_RZ, nan0, dcolor = _check(
        coll=coll,
        key=key,
        observation_directions=observation_directions,
        key_mesh=key_mesh,
        res_RZ=res_RZ,
        nan0=nan0,
        dcolor=dcolor,
    )

    # -----------
    # compute
    # -----------

    dpoly, ndet, extent = _compute(
        coll=coll,
        key=key,
        is_vos=is_vos,
        keym=keym,
        res_RZ=res_RZ,
        dcolor=dcolor,
    )

    # ------------
    # nan0

    if nan0 is True:
        ndet[ndet == 0] = np.nan

    # -----------
    # plot
    # -----------

    dax = _plot(
        coll=coll,
        key=key,
        # data
        is_vos=is_vos,
        ndet=ndet,
        extent=extent,
        dpoly=dpoly,
        # plotting options
        config=config,
        dax=dax,
        fs=fs,
        dmargin=dmargin,
        tit=tit,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
    )

    return dpoly, ndet, dax


# ################################################################
# ################################################################
#                           Check
# ################################################################


def _check(
    coll=None,
    key=None,
    observation_directions=None,
    key_mesh=None,
    res_RZ=None,
    nan0=None,
    dcolor=None,
):

    # -----------
    # key
    # -----------

    lok_vos = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if all([
            v1.get('dvos', {}).get('keym') is not None
            for v1 in v0['doptics'].values()
        ])
    ]
    lok = [
        k0 for k0 in coll.dobj.get('diagnostic', {}).keys()
        if k0 not in lok_vos
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok + lok_vos,
    )

    lcam = coll.dobj['diagnostic'][key]['camera']
    is_vos = key in lok_vos

    # is 3d ?
    doptics = coll.dobj['diagnostic'][key]['doptics']
    # is_3d = doptics[lcam[0]]['dvos'].get('indr_3d') is not None

    # ------------------------
    # observation_directions
    # ------------------------

    obsdef = False
    observation_directions = ds._generic_check._check_var(
        observation_directions, 'observation_directions',
        types=bool,
        default=obsdef,
    )

    # -------------
    # key mesh
    # -------------

    doptics = coll.dobj['diagnostic'][key]['doptics']
    if is_vos:

        # key_mesh unicity
        lmesh = set([v0['dvos']['keym'] for v0 in doptics.values()])

        # res unicity
        lres_RZ = set([tuple(v0['dvos']['res_RZ']) for v0 in doptics.values()])

        if len(lmesh) != 1 or len(lres_RZ) != 1:
            msg = (
                "Non-unique mesh or res_RZ for vos of diag '{key}':\n"

            )
            raise Exception(msg)

        keym = list(lmesh)[0]
        res_RZ = list(list(lres_RZ)[0])

    else:
        pass

    # -----------
    # nan0 => set 0 to nan
    # -----------

    nan0 = ds._generic_check._check_var(
        nan0, 'nan0',
        types=bool,
        default=True,
    )

    # -----------
    # dcolor
    # -----------

    if dcolor is None:
        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']
        dcolor = {k0: colors[ii%len(colors)] for ii, k0 in enumerate(lcam)}

    if mcolors.is_color_like(dcolor):
        dcolor = {k0: dcolor for k0 in lcam}

    c0 = (
        isinstance(dcolor, dict)
        and all([
            mcolors.is_color_like(dcolor.get(k0))
            for k0 in lcam
        ])
    )
    if not c0:
        lstr = [f"\t- {k0}: rgb or rgba color" for k0 in lcam]
        msg = (
            f"Arg dcolor must be a dict of (camera: color) for diag '{key}':\n"
            + "\n".join(lstr)
            + f"\nProvided:\n{dcolor}\n"
        )
        raise Exception(msg)

    # ----------
    # add alpha

    for k0, v0 in dcolor.items():
        if isinstance(v0, tuple) and len(v0) == 4:
            alpha = v0[-1]
        else:
            alpha = 0.5
        dcolor[k0] = mcolors.to_rgba(v0, alpha=alpha)

    return key, is_vos, observation_directions, keym, res_RZ, nan0, dcolor


# ################################################################
# ################################################################
#                           Compute
# ################################################################


def _compute(
    coll=None,
    key=None,
    is_vos=None,
    keym=None,
    res_RZ=None,
    dcolor=None,
):

    # ---------------
    # prepare
    # ---------------

    doptics = coll.dobj['diagnostic'][key]['doptics']

    # ---------------
    # mesh sampling
    # ---------------

    # ----------------
    # initial sampling

    dsamp = coll.get_sample_mesh(
        key=keym,
        res=res_RZ,
        mode='abs',
        grid=False,
        store=False,
    )

    R = dsamp['x0']['data']
    Z = dsamp['x1']['data']

    # --------------------
    # derive shape, extent

    # shape
    shape = (R.size, Z.size)

    # extent
    extent = (
        R[0] - 0.5*(R[1] - R[0]),
        R[-1] + 0.5*(R[-1] - R[-2]),
        Z[0] - 0.5*(Z[1] - Z[0]),
        Z[-1] + 0.5*(Z[-1] - Z[-2]),
    )

    # ---------------
    # ndet for vos
    # ---------------

    ndet = np.zeros(shape, dtype=float)

    if is_vos:
        for kcam, v0 in doptics.items():

            kindr, kindz = v0['dvos']['ind_cross']

            shape_poly = coll.ddata[kindr]['data'].shape
            lind = [range(ss) for ss in shape_poly[:-1]]

            sli0 = np.asarray(list(shape_poly[:-1]) + [slice(None)])
            sli = np.asarray(list(shape_poly[:-1]) + [slice(None)])

            for ii, ij in enumerate(itt.product(*lind)):

                # get rid of -1
                sli0[:-1] = ij
                sli[-1] = coll.ddata[kindr]['data'][tuple(sli0)] >= 0

                # get indices to incremente
                sli[:-1] = ij
                ind = (
                    coll.ddata[kindr]['data'][tuple(sli)],
                    coll.ddata[kindz]['data'][tuple(sli)],
                )

                ndet[ind] += 1

    # -----------------
    # ndet for non-vos
    # -----------------

    else:
        raise NotImplementedError()

    # ---------------
    # dpoly
    # ---------------

    dpoly = {}
    for ii, (kcam, v0) in enumerate(doptics.items()):

        # concatenate all vos
        pr, pz = _vos_utils._get_overall_polygons(
            coll=coll,
            doptics=doptics,
            key_cam=kcam,
            poly='pcross',
            convexHull=False,
        )

        # store
        dpoly[kcam] = {
            'pr': pr,
            'pz': pz,
            'color': dcolor[kcam],
        }

    return dpoly, ndet, extent


# ################################################################
# ################################################################
#                           plot
# ################################################################


def _plot(
    coll=None,
    key=None,
    # data
    ndet=None,
    extent=None,
    dpoly=None,
    # plotting options
    config=None,
    is_vos=None,
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
    cmap=None,
    vmin=None,
    vmax=None,
):

    # ----------------
    # check input
    # ----------------

    # tit
    if tit is not False:
        titdef = f"geometrical coverage of diag '{key}'"
        tit = ds._generic_check._check_var(
            tit, 'tit',
            types=str,
            default=titdef,
        )

    if cmap is None:
        cmap = plt.cm.viridis   # Greys

    if vmin is None:
        vmin = 0

    if vmax is None:
        vmax = np.nanmax(ndet)

    # ----------------
    # prepare data
    # ----------------

    # directions of observation



    # ----------------
    # prepare figure
    # ----------------

    if dax is None:
        if fs is None:
            fs = (14, 7)

        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.95,
                'bottom': 0.06, 'top': 0.90,
                'hspace': 0.20, 'wspace': 0.40,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=16, nrows=1, **dmargin)

        # ax0 = spans
        ax0 = fig.add_subplot(gs[0, :6], aspect='equal', adjustable='box')
        ax0.set_ylabel('Z (m)', size=12, fontweight='bold')
        ax0.set_xlabel('R (m)', size=12, fontweight='bold')
        ax0.set_title("spans", size=14, fontweight='bold')

        # ax1 = nb of detectors
        ax1 = fig.add_subplot(gs[0, 9:-1], sharex=ax0, sharey=ax0)
        ax1.set_ylabel('Z (m)', size=12, fontweight='bold')
        ax1.set_xlabel('R (m)', size=12, fontweight='bold')
        ax1.set_title("nb. of detectors", size=14, fontweight='bold')

        # colorbar
        cax = fig.add_subplot(gs[0, -1], frameon=False)

        dax = {
            'span': {'handle': ax0, 'type': 'span'},
            'ndet': {'handle': ax1, 'type': 'ndet'},
            'cax': {'handle': cax, 'type': 'span_colorbar'},
        }

    # --------------------
    # check / format dax

    dax = ds._generic_check._check_dax(dax)
    fig = list(dax.values())[0]['handle'].figure

    # ---------------
    # plot ndet
    # ---------------

    ktype = 'ndet'
    lax = [v0['handle'] for v0 in dax.values() if ktype in v0['type']]
    for ax in lax:

        # plot
        im = ax.imshow(
            ndet.T,
            extent=extent,
            origin='lower',
            interpolation='bilinear',
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )

        # colorbar
        ktype2 = 'span_colorbar'
        lcax = [v0['handle'] for v0 in dax.values() if ktype2 in v0['type']]
        if len(lcax) == 0:
            plt.colorbar(im)
        else:
            for cax in lcax:
                plt.colorbar(im, cax=cax)
                # cax.set_ylabel("nb. of detectors", size=14, fontweight='bold')

    # ---------------
    # plot spans
    # ---------------

    ktype = 'span'
    lax = [v0['handle'] for v0 in dax.values() if ktype in v0['type']]
    for ax in lax:

        for k0, v0 in dpoly.items():
            ax.fill(
                v0['pr'],
                v0['pz'],
                fc=v0['color'],
                label=k0,
            )

        ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left', fontsize=12)

    # --------------
    # add config
    # --------------

    if config is not None:
        for ktype in ['span', 'ndet']:
            lax = [v0['handle'] for v0 in dax.values() if ktype in v0['type']]
            for ax in lax:
                config.plot(lax=ax, proj='cross', dLeg=False)

    # --------------
    # add tit
    # --------------

    if tit is not False:
        fig.suptitle(tit, size=14, fontweight='bold')

    return dax



# -*- coding: utf-8 -*-

# Builtin
import warnings

# Common
import numpy as np
# import scipy.signal as scpsig
# import scipy.interpolate as scpinterp
# import scipy.linalg as scplin
# import scipy.stats as scpstats
import datastock as ds


#############################################
#############################################
#       ref indices propagation
#############################################


def _get_index_from_data(data=None, data_pick=None, monot=None):

    if isinstance(data, str) and data == 'index':
        indices = np.round(data_pick)

    elif monot:

        bins = np.r_[
            data[0] - 0.5*(data[1] - data[0]),
            0.5*(data[1:] + data[:-1]),
            data[-1] + 0.5*(data[-1] - data[-2]),
        ]
        indices = np.digitize(data_pick, bins, right=False)

        # corrections
        if data[1] > data[0]:
            indices[data_pick < bins[0]] = 0
            indices[data_pick > bins[-1]] = data.size-1
        else:
            indices[data_pick < bins[-1]] = data.size-1
            indices[data_pick > bins[0]] = 0

    elif data.ndim == 1:

        indices = np.array([
            np.nanargmin(np.abs(data - dd))
            for dd in data_pick
        ])

    else:
        msg = "Non-handled case yet"
        raise Exception(msg)

    return indices


def propagate_indices_per_ref(
    ref=None,
    lref=None,
    dref=None,
    ddata=None,
    ldata=None,
    param=None,
    value=None,
    lparam_data=None,
):

    # ------------
    # check inputs

    # ref
    ref = ds._generic_check._check_var(
        ref, 'ref',
        types=str,
        allowed=sorted(dref.keys())
    )

    ind = dref[ref].get('indices')
    if ind is None:
        return

    # lref
    if isinstance(lref, str):
        lref = [lref]
    lref = ds._generic_check._check_var_iter(
        lref, 'lref',
        types_iter=str,
        allowed=sorted(dref.keys()),
    )

    # param vs ldata
    if ldata is None:
        param = ds._generic_check._check_var(
            param, 'param',
            default='index',
            allowed=['index'] + lparam_data,
        )

    else:

        # check ldata length
        if len(ldata) != len(lref) + 1:
            msg = (
                "Arg ldata must contain one data key for ref + for each lref\n"
                f"\t- Provided: {ldata}"
            )
            raise Exception(msg)

        # check content
        dout = {
            rr: (ii, ldata[ii], dref[rr]['ldata_monot'])
            for ii, rr in enumerate([ref] + lref)
            if not (
                ldata[ii] in dref[rr]['ldata_monot']
                or ldata[ii] == 'index'
            )
        }
        if len(dout) > 0:
            lstr = [
                f"\t- {rr} ({vv[0]}): '{vv[1]}' vs {vv[2]}"
                for rr, vv in dout.items()
            ]
            msg = (
                "Provided ldata are not suitable:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

    # ---------
    # propagate

    if param == 'index':
        for rr in lref:
            dref[rr]['indices'] = dref[ref]['indices'] % dref[rr]['size']

    else:

        if ldata is None:
            # For ref, pick data
            ref_data = dref[ref]['ldata_monot']
            if len(ref_data) > 1:
                ref_data = [k0 for k0 in ref_data if ddata[k0][param] == value]
            if len(ref_data) != 1:
                msg = (
                    f"No / too many monotonous data for ref {ref}:\n"
                    f"\t- param / value: {param} / {value}\n"
                    f"\t- found: {ref_data}"
                )
                raise Exception(msg)
            ref_data = ref_data[0]

            # For each ref in lref, get list of matching data
            drdata = {
                rr: [
                    k0 for k0 in dref[rr]['ldata_monot']
                    if ddata[k0][param] == ddata[ref_data][param]
                ]
                for rr in [ref] + lref
            }

            # Raise exception if not unique
            dout = {rr: len(drdata[rr]) for rr in lref if len(drdata[rr]) != 1}
            if len(dout) > 0:
                lstr = [
                    f"\t- {rr}: {vv} matching monotonous data"
                    for rr, vv in dout.items()
                ]
                msg = (
                    "The following ref in lref have no/several matching data "
                    f"for param {param}:\n"
                    "\n".join(lstr)
                )
                raise Exception(msg)
            ldata = [drdata[rr][0] for rr in [ref] + lref]

        # propagate according to data (nearest neighbourg)
        if ldata[0] == 'index':
            dataref = None
        else:
            dataref = ddata[ldata[0]]['data']
            dataref = dataref[dref[ref]['indices']]
        for ii, rr in enumerate(lref):
            if ldata[ii+1] == 'index':
                data = dref[rr]['indices']
            else:
                data = ddata[ldata[ii+1]]['data']
            dref[rr]['indices'] = _get_index_from_data(
                data=dataref,
                data_pick=data,
                monot=True,
            )


#############################################
#############################################
#############################################
#   common reference (e.g.: time vectors)
#############################################
#############################################


def _get_unique_ref_dind(
    dd=None, dd_name='data', group='time',
    lkey=None,
    return_all=None,
):
    """ Typically used to get a common (intersection) time vector

    Returns a time vector that contains all time points from all data
    Also return a dict of indices to easily remap each time vector to tall
            such that tall[ind] => t (with nearest approximation)
            so ind contains indices of t such that tall falls in t

    dind[k0] = np.searchsorted(tbin, tall)

    """

    if return_all is None:
        return_all = False
    if isinstance(lkey, str):
        lkey = [lkey]

    c0 = (
        isinstance(lkey, list)
        and all([k0 in dd.keys() for k0 in lkey])
    )
    if not c0:
        msg = "Non-valid keys provided:\n{}".format(
            [kk for kk in lkey if kk not in dd.keys()]
        )
        raise Exception(msg)

    # Only keep keys with group
    lkey = [kk for kk in lkey if group in dd[kk]['group']]
    dind = dict.fromkeys(lkey)
    if len(lkey) == 0:
        if return_all is True:
            return None, None, 1, dind
        else:
            return None, dind

    # get list of ref from desired group (e.g.: time vectors id)
    did = {
        k0: [
            id_ for id_ in dd[k0]['ref'] if dd[id_]['group'] == (group,)
        ][0]
        for k0 in lkey
    }
    lidu = list(set([vv for vv in did.values()]))

    # Create common time base
    if len(lidu) == 1:
        tall = dd[lidu[0]]['data']

    else:
        tall = np.unique([dd[kt]['data'] for kt in lidu])

    # Get dict if indices for fast mapping each t to tall
    for k0 in lkey:
        dind[k0] = np.searchsorted(
            0.5*(dd[did[k0]]['data'][1:] + dd[did[k0]]['data'][:-1]),
            tall,
        )

    # Other output
    if return_all is True:
        tbinall = 0.5*(tall[1:] + tall[:-1])
        ntall = tall.size
        return tall, tbinall, ntall, dind
    else:
        return tall, dind


def _get_indtmult(
    dd=None, dd_name='data', group='time',
    idquant=None, idref1d=None, idref2d=None,
):

    # Get time vectors and bins
    idtq = [
        id_ for id_ in dd[idquant]['ref'] if dd[id_]['group'] == (group,)
    ][0]
    tq = dd[idtq]['data']
    tbinq = 0.5*(tq[1:] + tq[:-1])

    # Get time vectors for ref1d and ref2d if any
    if idref1d is not None:
        idtr1 = [
            id_ for id_ in dd[idref1d]['ref'] if dd[id_]['group'] == (group,)
        ][0]
        tr1 = dd[idtr1]['data']
        tbinr1 = 0.5*(tr1[1:] + tr1[:-1])

    if idref2d is not None and idref2d != idref1d:
        idtr2 = [
            id_ for id_ in dd[idref2d]['ref'] if dd[id_]['group'] == (group,)
        ][0]
        tr2 = dd[idtr2]['data']
        tbinr2 = 0.5*(tr2[1:] + tr2[:-1])

    # Get tbinall and tall
    if idref1d is None:
        tbinall = tbinq
        tall = tq
    else:
        if idref2d is None:
            tbinall = np.unique(np.r_[tbinq, tbinr1])
        else:
            tbinall = np.unique(np.r_[tbinq, tbinr1, tbinr2])
        tall = np.r_[tbinall[0] - 0.5*(tbinall[1]-tbinall[0]),
                     0.5*(tbinall[1:]+tbinall[:-1]),
                     tbinall[-1] + 0.5*(tbinall[-1]-tbinall[-2])]

    # Get indtqr1r2 (tall with respect to tq, tr1, tr2)
    indtq, indtr1, indtr2 = None, None, None
    if tbinq.size > 0:
        indtq = np.digitize(tall, tbinq)
    else:
        indtq = np.r_[0]

    if idref1d is None:
        assert np.all(indtq == np.arange(0, tall.size))
    if idref1d is not None:
        if tbinr1.size > 0:
            indtr1 = np.digitize(tall, tbinr1)
        else:
            indtr1 = np.r_[0]

    if idref2d is not None:
        if tbinr2.size > 0:
            indtr2 = np.digitize(tall, tbinr2)
        else:
            indtr2 = np.r_[0]

    ntall = tall.size
    return tall, tbinall, ntall, indtq, indtr1, indtr2


def _get_indtu(
    t=None,
    tall=None,
    tbinall=None,
    indtq=None, indtr1=None, indtr2=None,
):
    """ Return relevent time indices

    Typically:
        indt = np.searchsorted(tbinall, t)
        t[indt] => tall
        indtu = np.unique(indt)

        tq[indtq[ii]] = tall[ii]
        tr1[indtr1[ii]] = tall[ii]
        tr2[indtr2[ii]] = tall[ii]

    """

    # Get indt (t with respect to tbinall)
    if tall is not None and t is not None:
        if len(t) == len(tall) and np.allclose(t, tall):
            indt = np.arange(0, tall.size)
            indtu = indt
        else:
            indt = np.searchsorted(tbinall, t)
            indtu = np.unique(indt)
    elif tall is not None and t is None:
        indt = np.arange(0, tall.size)
        indtu = indt
    elif tall is None and t is not None:
        indtu = np.array([0], dtype=int)
        indt = np.zeros((t.size,), dtype=int)
    else:
        indtu = np.array([0], dtype=int)
        indt = indtu

    # indices for ref1d and ref2d
    if indtq is None:
        indtq = np.zeros((tall.size,), dtype=int)
    if indtr1 is None:
        indtr1 = np.zeros((tall.size,), dtype=int)
    if indtr2 is None:
        indtr2 = np.zeros((tall.size,), dtype=int)

    # Update
    tall = tall[indtu]
    ntall = indtu.size

    return tall, ntall, indt, indtu, indtq, indtr1, indtr2


def get_tcommon(self, lq, prefer='finer'):
    """ Check if common t, else choose according to prefer

    By default, prefer the finer time resolution

    """
    if type(lq) is str:
        lq = [lq]
    t = []
    for qq in lq:
        ltr = [kk for kk in self._ddata[qq]['depend']
               if self._dindref[kk]['group'] == 'time']
        assert len(ltr) <= 1
        if len(ltr) > 0 and ltr[0] not in t:
            t.append(ltr[0])
    assert len(t) >= 1
    if len(t) > 1:
        dt = [np.nanmean(np.diff(self._ddata[tt]['data'])) for tt in t]
        if prefer == 'finer':
            ind = np.argmin(dt)
        else:
            ind = np.argmax(dt)
    else:
        ind = 0
    return t[ind], t


def _get_tcom(
    idquant=None, idref1d=None,
    idref2d=None, idq2dR=None,
    dd=None, group=None,
):
    if idquant is not None:
        out = _get_unique_ref_dind(
            dd=dd, group=group,
            lkey=[idquant, idref1d, idref2d],
            return_all=True,
        )
    else:
        out = _get_unique_ref_dind(
            dd=dd, group=group,
            lkey=[idq2dR],
            return_all=True,
        )
    return out


#############################################
#############################################
#############################################
#       1d fits
#############################################
#############################################

def fit_1d(data, x=None, axis=None, Type=None, func=None,
           dTypes=None, **kwdargs):

    lc = [Type is not None, func is not None]
    assert np.sum(lc) == 1

    if lc[0]:

        # Pre-defined models dict
        # ------------------------

        _DTYPES = {'staircase': _fit1d_staircase}

        # Use a pre-defined model
        # ------------------------

        if dTypes is None:
            dTypes = _DTYPES

        if Type not in dTypes.keys():
            msg = "Chosen Type not available:\n"
            msg += "    - provided: {}\n".format(Type)
            msg += "    - Available: {}".format(list(dTypes.keys()))
            raise Exception(msg)

        dout = dTypes[Type](data, x=x, axis=axis, **kwdargs)

    else:

        # Use a user-provided model
        # -------------------------

        dout = func(data, x=x, axis=axis, **kwdargs)

    return dout


# -------------------------------------------
#       1d fit models
# -------------------------------------------


def _fit1d_staircase(data, x=None, axis=None):
    """ Model data as a staircase (ramps + plateaux)

    Return a the fitted parameters as a dict:
        {'plateaux': {'plateaux': {'Dt': (2,N) np.ndarray
                                   '':
                                    }}
        - to be discussed.... ?

    """
    pass



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import scipy.stats as scpstats
import datastock as ds


# ###############################################################
# ###############################################################
#                   Binning of 2d spectro data
# ###############################################################


def binned(
    coll=None,
    key_diag=None,
    key_cam=None,
    # data to be binned
    data=None,
    statistic=None,
    # binning dimension
    bins0=None,
    bins1=None,
    bin_data0=None,
    bin_data1=None,
    # store
    store=None,
    # plotting
    plot=None,
    ax=None,
):

    # ------------
    # check

    (
        key_diag, key_cam, spectro, is2d,
        ddata, dref, units,
        dbins0, dbins1,
        statistic, store, plot,
    ) = _binned_check(**locals())

    # ----------
    # compute

    dout = {}
    for ii, kcam in enumerate(ddata.keys()):

        dati = ddata[kcam]
        bind = dbins0['ddata'][kcam]

        # binning index
        if dbins1 is None:

            for kcam in key_cam:
                dout[kcam] = coll.binning(
                    keys=data,
                    ref_key=dbins0['key'][''],
                    bins=dbins0['dedges'][kcam],
                    verb=False,
                    store=False,
                    returnas=None,
                    key_store=None,
                )

                import pdb; pdb.et_trace()      # DB

            iok = np.isfinite(dati) & np.isfinite(bind)
            out = scpstats.binned_statistic(
                bind[iok],
                dati[iok],
                statistic=statistic,
                bins=dbins0['dedges'][kcam],
            )

        else:
            out = scpstats.binned_statistic_2d(
                bind[iok],
                dati[iok],
                statistic=statistic,
                bins=[dbins0['dedges'][kcam], dbins1['dedges'][kcam]],
            )

        import pdb; pdb.et_trace()      # DB

        # fill dict
        # dout[kcam] = {
        #     'data': out,
        #     'units': units,
        #     'bins': 0.5*(dbin_edges[kcam][1:] + dbin_edges[kcam][:-1]),
        #     'bin_units': bin_dd[kcam]['units'],
        # }

    # ----------
    # plot

    if plot is True:
        ax = _spectro2d_binned_plot(
            key_diag=key_diag,
            # data to be binned
            data=None,
            # binning dimension
            bin_data=None,
            dout=dout,
            # plotting
            ax=ax,
        )

    # ----------
    # return

    if plot is True:
        return dout, ax
    else:
        return dout


def _binned_check(
    coll=None,
    key_diag=None,
    key_cam=None,
    # data to be binned
    data=None,
    # binning dimension
    bins0=None,
    bins1=None,
    bin_data0=None,
    bin_data1=None,
    statistic=None,
    # store
    store=None,
    # plotting
    plot=None,
    # others
    **kwdargs,
):

    # --------
    # key_diag

    key_diag, key_cam = coll.get_diagnostic_cam(key=key_diag, key_cam=key_cam)
    spectro = coll.dobj['diagnostic'][key_diag]['spectro']
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']

    # ----------
    # data

    (
        ddata, dref,
        units, _, _,
    ) = coll.get_diagnostic_data(key_diag=key_diag, key_cam=key_cam, data=data)
    key_cam = list(ddata.keys())

    dref_cam = {k0: coll.dobj['camera'][k0]['dgeom']['ref'] for k0 in key_cam}

    # --------------
    # bin_data, bins

    dbins0 = _check_bins(
        coll=coll,
        dref=dref,
        key_diag=key_diag,
        dref_cam=dref_cam,
        key_cam=key_cam,
        bin_data=bin_data0,
        bins=bins0,
    )

    if bin_data1 is not None:
        dbins1 = _check_bins(
            coll=coll,
            dref=dref,
            dref_cam=dref_cam,
            key_diag=key_diag,
            key_cam=key_cam,
            bin_data=bin_data1,
            bins=bins1,
        )
    else:
        dbins1 = None

    # --------
    # statistic

    statistic = ds._generic_check._check_var(
        statistic, 'statistic',
        types=str,
        default='sum',
    )

    # --------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=False,
    )

    # --------
    # plot

    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    return (
        key_diag, key_cam, spectro, is2d,
        ddata, dref, units,
        dbins0, dbins1,
        statistic, store, plot,
    )



def _check_bins(
    coll=None,
    ddata=None,
    bin_data=None,
    bins=None,
    bin_units=None,
    
    # dref=None,
    # dref_cam=None,
    # key_diag=None,
    # key_cam=None,
    # bin_data=None,
    # bins=None,
    # bin_units=None,
    # # if bsplines
    # safety_ratio=None,
    # strict=None,
    # deg=None,
):

    # --------------
    # options

    # check
    strict = _generic_check._check_var(
        strict, 'strict',
        types=bool,
        default=True,
    )

    # check
    safety_ratio = _generic_check._check_var(
        safety_ratio, 'safety_ratio',
        types=(int, float),
        default=1.5,
        sign='>0.'
    )

    # --------------
    # bins

    if bins is None:
        bins = 100

    if np.isscalar(bins):
        bins = int(bins)

    else:
        bins = _generic_check._check_flat1d_array(
            bins, 'bins',
            dtype=float,
            unique=True,
            can_be_None=False,
        )

    # -------------
    # bin data

    if isinstance(bin_data, str):

        # lquant = ['etendue', 'amin', 'amax']  # 'los'
        # lcomp = ['length', 'tangency radius', 'alpha']
        # llamb = ['lamb', 'lambmin', 'lambmax', 'dlamb', 'res']
        # lok_fixed = ['x0', 'x1'] + lquant + lcomp + llamb

        lok_static = []
        lok_var = []
        for k0 in coll.dobj['diagnostic'][key_diag]['signal']:
            cams = coll.dobj['synth sig'][k0]['camera']
            ref = coll.ddata[coll.dobj['synth sig'][k0]['data'][0]]['ref']
            if ref == dref_cam[cams[0]]:
                lok_sig_static.append(k0)
            elif ref == dref[cams[0]]:
                lok_sig_var.append(k0)

        bin_key = ds._generic_check._check_var(
            bin_data, 'bin_data',
            types=str,
            allowed=lok_fixed + lok_sig_static + lok_sig_var,
        )
        bin_data = coll.ddata[bin_key]['data']
        bin_ref = coll.ddata[bin_key]['ref']
        bin_units = coll.ddata[bin_key]['units']

        variable = bin_data in lok_sig_var

    elif isinstance(bin_data, np.ndarray):
        bin_key = None
        shape = tuple([ss for ss in data_shape if ss in bin_data.shape])
        if bin_data.shape != shape:
            msg = "Arg bin_data must have "
            raise Exception(msg)

    elif bin_data is None:
        hasref, ref, bin_data, val, dkeys = coll.get_ref_vector_common(
            keys=keys,
            strategy=ref_vector_strategy,
        )
        if bin_data is None:
            msg = (
                f"No matching ref vector found for:\n"
                f"\t- keys: {keys}\n"
                f"\t- hasref: {hasref}\n"
                f"\t- ref: {ref}\n"
                f"\t- ddata['{keys[0]}']['ref'] = {coll.ddata[keys[0]]['ref']} "
            )
            raise Exception(msg)

    else:
        msg = f"Invalid bin_data:\n{bin_data}"
        raise Exception(msg)

    # ----------------
    # get axis

    if axis is None:
        if bin_key is None:
            axis = np.array([
                ii for ii, ss in enumerate(data_shape)
                if ss in bin_data.shape
            ])
        else:
            axis = np.array([
                ii for ii, rr in enumerate(data_ref)
                if rr in bin_ref
            ])

    axis = ds._generic_check._check_flat1d_array(
        axis, 'axis',
        dtype=int,
        unique=True,
        can_be_None=False,
        sign='>=0',
    )

    if np.any(axis > len(data_shape)-1):
        msg = f"axis too large\n{axis}"
        raise Exception(msg)

    if np.any(np.diff(axis) > 1):
        msg = f"axis must be adjacent indices!\n{axis}"
        raise Exception(msg)

    # --------------
    # bins

    # bins
    if isinstance(bins, int):
        bin_min = np.nanmin(bin_data)
        bin_max = np.nanmax(bin_data)
        bin_edges = np.linspace(bin_min, bin_max, bins + 1)

    else:
        bin_edges = np.r_[
            bins[0] - 0.5*(bins[1] - bins[0]),
            0.5*(bins[1:] + bins[:-1]),
            bins[-1] + 0.5*(bins[-1] - bins[-2]),
        ]

    # ----------
    # bin method

    if bin_data.ndim == 1:
        dv = np.abs(np.diff(bin_data))
        dv = np.append(dv, dv[-1])
        dvmean = np.mean(dv) + np.std(dv)

        if strict is True:
            lim = safety_ratio*dvmean
            if not np.mean(np.diff(bin_edges)) > lim:
                msg = (
                    f"Uncertain binning for '{sorted(keys)}', ref vect '{ref_key}':\n"
                    f"Binning steps ({db}) are < {safety_ratio}*ref ({lim}) vector step"
                )
                raise Exception(msg)

            else:
                npts = None

        else:
            npts = (deg + 3) * max(1, dvmean / db)

    # ----------
    # dbins

    dbins = {
        'key': bin_key,
        'bins': bins,
        'edges': edges,
        'data': bin_data,
        'ref': bin_ref,
        'units': bin_units,
        'axis': axis,
    }

    return dbins, npts




def _spectro2d_binned_plot(
    key=None,
    key_cam=None,
    # data to be binned
    data=None,
    # binning dimension
    bin_data=None,
    dout=None,
    # plotting
    ax=None,
):

    # ------------
    #



    return ax



# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    Magnetic field lines
'''

import warnings
import traceback

msg, err = None, None
try:
    import imas
except Exception as err:
    if str(err) == 'imas not available':
        msg = ""
        msg += "\n\nIMAS python API issue\n"
        msg += "imas could not be imported into tofu ('import imas' failed):\n"
        msg += "  - it may not be installed (optional dependency)\n"
        msg += "  - or you not have loaded the good working environment\n\n"
        msg += "    => the optional sub-package tofu.mag is not usable\n"
    else:
        msg = str(traceback.format_exc())
        msg += "\n\n    => the optional sub-package tofu.mag is not usable\n"
    raise Exception(msg)

try:
    import pywed
except Exception:
    msg = "pywed not available => no tofu.mag"
    raise Exception(msg)

try:
    from tofu.mag.magFieldLines import *
except Exception:
    try:
        from .magFieldLines import *
    except Exception:
        msg = "Could not import"
        raise Exception(msg)

del warnings, traceback, pywed, imas
del msg, err

__all__ = ['MagFieldLines']



"""
This module contains tests for tofu.geom in its structured version
"""

# External modules
import os
import itertools as itt
import numpy as np
import matplotlib.pyplot as plt
import warnings as warn

# Importing package tofu.gem
import tofu as tf
from tofu import __version__
import tofu.defaults as tfd
import tofu.utils as tfu
import tofu.geom as tfg

import tofu.spectro._rockingcurve_def as _rockingcurve_def

_here = os.path.abspath(os.path.dirname(__file__))
_PATH_DATA = os.path.join(_here, 'test_data')
_PFE_DET = os.path.join(_PATH_DATA, 'det37_CTVD_incC4_New.npz')

VerbHead = 'tofu.geom.test_04_core_optics'
keyVers = 'Vers'
_Exp = 'WEST'




#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module(module):
    print("")   # this is to get a newline after the dots
    lf = os.listdir(_here)
    lf = [
        ff for ff in lf
        if all([s in ff for s in ['TFG_', _Exp, '.npz']])
    ]
    lF = []
    for f in lf:
        ff = f.split('_')
        v = [fff[len(keyVers):] for fff in ff
             if fff[:len(keyVers)] == keyVers]
        msg = f + "\n    "+str(ff) + "\n    " + str(v)
        assert len(v) == 1, msg
        v = v[0]
        if '.npz' in v:
            v = v[:v.index('.npz')]
        # print(v, __version__)
        if v != __version__:
            lF.append(f)
    if len(lF) > 0:
        print("Removing the following previous test files:")
        for f in lF:
            os.remove(os.path.join(_here, f))
        # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    lf = os.listdir(_here)
    lf = [
        ff for ff in lf
        if all([s in ff for s in ['TFG_', _Exp, '.npz']])
    ]
    lF = []
    for f in lf:
        ff = f.split('_')
        v = [fff[len(keyVers):] for fff in ff
             if fff[:len(keyVers)] == keyVers]
        msg = f + "\n    "+str(ff) + "\n    " + str(v)
        assert len(v) == 1, msg
        v = v[0]
        if '.npz' in v:
            v = v[:v.index('.npz')]
        # print(v, __version__)
        if v == __version__:
            lF.append(f)
    if len(lF) > 0:
        print("Removing the following test files:")
        for f in lF:
            os.remove(os.path.join(_here, f))


#######################################################
#
#   Crystal class
#
#######################################################

class Test01_Crystal(object):

    @classmethod
    def setup_class(cls, verb=False):
        # print ("")
        # print "--------- "+VerbHead+cls.__name__

        # Prepare input
        dgeom = {
            'Type': 'sph',
            'Typeoutline': 'rect',
            'summit': np.array([4.6497750e-01, -8.8277925e+00, 3.5125000e-03]),
            'center': np.array([1.560921, -6.31106476, 0.00729429]),
            'extenthalf': np.array([0.01457195, 0.01821494]),
            'rcurve': 2.745,
            'move': 'rotate_around_3daxis',
            'move_param': 0.022889993139905633,
            'move_kwdargs': {
                'axis': np.array([
                    [4.95e-01, -8.95e+00, -8.63e-02],
                    [-1.37e-04, -2.18e-03,  9.99e-01],
                ])
            }
        }
        dmat = {
            'formula': 'Quartz',
            'density': 2.6576,
            'symmetry': 'hexagonal',
            'lengths': np.array([4.9079e-10, 4.9079e-10, 5.3991e-10]),
            'angles': np.array([1.57079633, 1.57079633, 2.0943951]),
            'cut': np.array([1,  1, -2,  0]),
            'd': 2.4539499999999996e-10,
        }
        dbragg = {
            'lambref': 3.96e-10,
        }

        dmat1 = dict(dmat)
        dmat2 = dict(dmat)
        dmat3 = dict(dmat)

        # cryst1
        cryst1 = tfg.CrystalBragg(
            dgeom=dgeom,
            dmat=dmat1,
            dbragg=dbragg,
            Name='Cryst1',
            Diag='SpectrX2D',
            Exp='WEST',
        )

        # cryst2
        dmat2['alpha'] = 0.
        dmat2['beta'] = 0.
        cryst2 = tfg.CrystalBragg(
            dgeom=dgeom,
            dmat=dmat2,
            dbragg=dbragg,
            Name='Cryst2',
            Diag='SpectrX2D',
            Exp='WEST',
        )

        # cryst3
        dmat3['alpha'] = (3/60)*np.pi/180
        dmat3['beta'] = 0.
        cryst3 = tfg.CrystalBragg(
            dgeom=dgeom,
            dmat=dmat3,
            dbragg=dbragg,
            Name='Cryst3',
            Diag='SpectrX2D',
            Exp='WEST',
        )

        # cryst4
        dmat['alpha'] = (3/60)*np.pi/180
        dmat['beta'] = np.pi/1000.
        cryst4 = tfg.CrystalBragg(
            dgeom=dgeom,
            dmat=dmat,
            dbragg=dbragg,
            Name='Cryst4',
            Diag='SpectrX2D',
            Exp='WEST',
        )

        cls.dobj = {
            'cryst1': cryst1,
            'cryst2': cryst2,
            'cryst3': cryst3,
            'cryst4': cryst4,
        }

        cls.xi = 0.05*np.linspace(-1, 1, 100)
        cls.xj = 0.10*np.linspace(-1, 1, 200)

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self):
        # print ("TestUM:setup_method() before each test method")
        pass

    def teardown_method(self):
        # print ("TestUM:teardown_method() after each test method")
        pass

    # def test00_todo(self):
        # pass

    def test01_todict(self):
        for k0 in self.dobj.keys():
            dd = self.dobj[k0].to_dict()
            assert type(dd) is dict

    def test02_fromdict(self):
        for k0 in self.dobj.keys():
            dd = self.dobj[k0].to_dict()
            obj = tfg.CrystalBragg(fromdict=dd)
            assert isinstance(obj, self.dobj[k0].__class__)

    def test03_copy_equal(self):
        for k0 in self.dobj.keys():
            obj = self.dobj[k0].copy()
            assert obj == self.dobj[k0]
            assert not obj != self.dobj[k0]

    def test04_get_nbytes(self):
        for k0 in self.dobj.keys():
            nb, dnb = self.dobj[k0].get_nbytes()

    def test05_strip_nbytes(self, verb=False):
        lok = tfg.CrystalBragg._dstrip['allowed']
        nb = np.full((len(lok),), np.nan)
        for k0, obj in self.dobj.items():
            for ii in lok:
                obj.strip(ii)
                nb[ii] = obj.get_nbytes()[0]
            assert np.all(np.diff(nb) <= 0.)
            for ii in lok[::-1]:
                obj.strip(ii)

    def test06_set_move_None(self):
        pass

    def test07_rotate_copy(self):
        pass

    def test08_get_detector_ideal(self):
        for k0, obj in self.dobj.items():
            det0 = obj.get_detector_ideal(miscut=False)
            det1 = obj.get_detector_ideal(miscut=True)
            assert isinstance(det0, dict) and isinstance(det0, dict)
            lk = ['nout', 'ei']
            assert all([ss in det0.keys() for ss in lk])
            assert all([ss in det1.keys() for ss in lk])
            if k0 in ['cryst1', 'cryst2']:
                assert all([
                    np.allclose(det0[kk], det1[kk])
                    for kk in lk
                ])
            elif k0 in ['cryst3', 'cryst4']:
                assert not any([
                    np.allclose(det0[kk], det1[kk])
                    for kk in lk
                ])
                for k1, v1 in det0.items():
                    assert np.linalg.norm(v1 - det1[k1]) <= 0.01

    def test09_plot(self):
        ii = 0
        for k0, obj in self.dobj.items():
            det = obj.get_detector_ideal()
            det['outline'] = np.array([
                0.1*np.r_[-1, 1, 1, -1, -1],
                0.1*np.r_[-1, -1, 1, 1, -1],
            ])
            pts, vect = obj.get_rays_from_cryst(
                phi=np.pi, returnas='(pts, vect)',
            )
            dist = obj.get_rowland_dist_from_lambbragg()
            pts = pts + dist*np.r_[0.5, 1., 2][None, :]*vect[:, 0:1, 0]
            lamb = obj.dbragg['lambref'] + np.r_[-1, 0, 1, 2]*1-12
            dax = obj.plot(
                pts=pts,
                lamb=lamb,
                det=det,
                rays_color='pts' if ii % 2 == 0 else 'lamb',
            )
            ii += 1
        plt.close('all')

    def test10_get_lamb_avail_from_pts(self):
        for k0, obj in self.dobj.items():
            det = obj.get_detector_ideal()
            pts, vect = obj.get_rays_from_cryst(
                phi=-9*np.pi/10., returnas='(pts, vect)',
            )
            dist = obj.get_rowland_dist_from_lambbragg()
            pts = pts + dist*np.r_[0.5, 1., 2][None, :]*vect[:, :, 0]
            lamb, phi, dtheta, psi, xi, xj = obj.get_lamb_avail_from_pts(
                pts=pts, det=det,
            )
            pts = pts + np.r_[7.5][None, :]*vect[:, :, 0]
            lamb, phi, dtheta, psi, xi, xj = obj.get_lamb_avail_from_pts(
                pts=pts, det=det,
            )
            conf = tf.load_config('WEST-V0')
            pts, dv, ind, res_eff = conf.Ves.V1.get_sampleV(
                res=0.3,
                domain=[None, None, [-np.pi, -np.pi/2.]],
            )
            lamb, phi, dtheta, psi, xi, xj = obj.get_lamb_avail_from_pts(
                pts=pts, det=det, strict=True,
            )

    def test11_calc_johann_error(self):
        for k0, obj in self.dobj.items():
            det = obj.get_detector_ideal()
            err_lamb, err_phi, _, _, _ = obj.calc_johannerror(
                xi=self.xi,
                xj=self.xj,
                det=det,
            )

    def test12_plot_line_on_det_tracing(self):
        for k0, obj in self.dobj.items():
            det = obj.get_detector_ideal()
            det['outline'] = np.array([
                0.1*np.r_[-1, 1, 1, -1, -1],
                0.1*np.r_[-1, -1, 1, 1, -1],
            ])
            dax = obj.plot_line_on_det_tracing(
                det=det,
                crystal='Quartz_110',
                merge_rc_data=True,
                miscut=False,
                therm_exp=False,
                plot=False,
            )

    def test13_calc_meridional_sagittal_focus(self):
        derr = {}
        for k0, obj in self.dobj.items():
            out = obj.calc_meridional_sagittal_focus(
                miscut=False,
                verb=False,
            )
            c0 = round(out[0], ndigits=12) == round(out[2], ndigits=12)
            c1 = round(out[1], ndigits=12) == round(out[3], ndigits=12)
            if not c0:
                derr[k0] = f'Meridional ({out[0]} vs {out[2]})'
                if not c1:
                    derr[k0] += f' + Sagittal ({out[1]} vs {out[3]})'
                derr[k0] += ' focus wrong'
            elif not c1:
                derr[k0] = f'Sagittal ({out[1]} vs {out[3]}) focus wrong'

            if obj.dmat['alpha'] != 0.0:
                out = obj.calc_meridional_sagittal_focus(
                    miscut=True,
                    verb=False,
                )
                c0 = round(out[0], ndigits=12) != round(out[2], ndigits=12)
                c1 = round(out[1], ndigits=12) != round(out[3], ndigits=12)
                if not c0:
                    derr[k0] = f'Meridional ({out[0]} vs {out[2]})'
                    if not c1:
                        derr[k0] += f' + Sagittal ({out[1]} vs {out[3]})'
                    derr[k0] += ' focus wrong'
                elif not c1:
                    derr[k0] = f'Sagittal ({out[1]} vs {out[3]}) focus wrong'
        if len(derr) > 0:
            lstr = [f'\t- {k0}: {v0}' for k0, v0 in derr.items()]
            msg = (
                "The following crystals have wrong focus:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

    def test14_plot_focal_error_summed(self):
        det = dict(np.load(
            os.path.join(_PATH_DATA, 'det37_CTVD_incC4_New.npz'),
            allow_pickle=True,
            ))
        for k0, obj in self.dobj.items():
            out = obj.plot_focal_error_summed(
                dist_min=-0.02, dist_max=0.02, ndist=5,
                di_min=-0.02, di_max=0.02, ndi=5,
                xi=self.xi[::20], xj=self.xj[::20],
                miscut=False,
                det_ref=det,
                plot_dets=True,
            )
        plt.close('all')

    def test15_split(self):
        for ii, (k0, obj) in enumerate(self.dobj.items()):
            direction = None if ii == 0 else ('e1' if ii % 2 == 0 else 'e2')
            nb = None if ii == 0 else (2 if ii % 2 == 0 else 3)
            lcryst = obj.split(direction=direction, nb=nb)

    def test16_get_plasmadomain_at_lamb(self):

        # load useful objects
        conf0 = tf.load_config('WEST-V0')
        det = dict(np.load(_PFE_DET, allow_pickle=True))

        # test all crystals
        for ii, (k0, obj) in enumerate(self.dobj.items()):
            if ii % 2 == 0:
                plot_as = 'poly'
                xixj_lim = None
                domain = [None, None, [-4*np.pi/5., -np.pi/2.]]
                res = 0.05
            else:
                plot_as = 'pts'
                xixj_lim = [[-0.04, 0.04], 172.e-4*np.r_[-0.5, 0.5]]
                domain = [None, [-0.1, 0.1], [-4*np.pi/5., -np.pi/2.]]
                res = 0.02

            pts, lambok, dax = obj.get_plasmadomain_at_lamb(
                det=det,
                lamb=[3.94e-10, 4.e-10],
                res=res,
                config=conf0,
                domain=domain,
                xixj_lim=xixj_lim,
                plot_as=plot_as,
            )
        plt.close('all')

    def test17_saveload(self, verb=False):
        for k0, obj in self.dobj.items():
            obj.strip(-1)
            pfe = obj.save(verb=verb, return_pfe=True)
            obj2 = tf.load(pfe, verb=verb)
            msg = "Unequal saved / loaded objects !"
            assert obj2 == obj, msg
            # Just to check the loaded version works fine
            obj2.strip(0)
            os.remove(pfe)



# -*- coding: utf-8 -*-


import warnings


import numpy as np
import scipy.stats as scpstats
from matplotlib.path import Path
import matplotlib.pyplot as plt       # DB
import matplotlib.gridspec as gridspec
import Polygon as plg
import datastock as ds


from . import _class8_equivalent_apertures as _equivalent_apertures
from . import _class8_vos_utilities as _utilities
from . import _class8_reverse_ray_tracing as _reverse_rt
from . import _class8_vos as _vos


# ###############################################################
# ###############################################################
#                       Main
# ###############################################################


def compute_vos_nobin_at_lamb(
    coll=None,
    key_diag=None,
    key_cam=None,
    key_mesh=None,
    # wavelength
    lamb=None,
    # config
    config=None,
    # parameters
    res_RZ=None,
    res_phi=None,
    res_rock_curve=None,
    n0=None,
    n1=None,
    convexHull=None,
    # margins
    margin_poly=None,
    # options
    nmax_rays=None,
    # spectro-only
    rocking_curve_fw=None,
    rocking_curve_max=None,
    ## binning
    dobin=None,
    bin0=None,
    bin1=None,
    remove_raw=None,
    # bool
    visibility=None,
    convex=None,
    verb=None,
    debug=None,
    # plot
    plot=None,
    pix0=None,
    pix1=None,
    tit=None,
):

    # ------------
    # check inputs

    (
        # standard vos
        key_diag,
        key_mesh,
        spectro,
        is2d,
        doptics,
        dcompute,
        res_RZ,
        res_phi,
        convexHull,
        visibility,
        verb,
        debug,
        # specific
        dobin,
        bin0,
        bin1,
        remove_raw,
        lamb,
        nmax_rays,
        plot,
        pix0,
        pix1,
    ) = _check(**locals())

    # ------------
    # sample mesh

    dsamp = coll.get_sample_mesh(
        key=key_mesh,
        res=res_RZ,
        mode='abs',
        grid=True,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    sh = dsamp['x0']['data'].shape
    x0f = dsamp['x0']['data'].ravel()
    x1f = dsamp['x1']['data'].ravel()

    sh1 = tuple([ss + 2 for ss in sh])
    bool_cross = np.zeros(sh1, dtype=bool)

    x0u = dsamp['x0']['data'][:, 0]
    x1u = dsamp['x1']['data'][0, :]
    x0l = np.r_[x0u[0] - (x0u[1] - x0u[0]), x0u, x0u[-1] + (x0u[-1] - x0u[-2])]
    x1l = np.r_[x1u[0] - (x1u[1] - x1u[0]), x1u, x1u[-1] + (x1u[-1] - x1u[-2])]
    x0l = np.repeat(x0l[:, None], x1l.size, axis=1)
    x1l = np.repeat(x1l[None, :], x0l.shape[0], axis=0)

    dx0 = x0u[1] - x0u[0]
    dx1 = x1u[1] - x1u[0]

    # --------------
    # prepare optics

    doptics = coll._dobj['diagnostic'][key_diag]['doptics']

    # --------------
    # prepare optics

    dout = {}
    for k0 in dcompute.keys():

            # ------------------
            # call relevant func

            dout[k0] = _vos_nobin_at_lamb(
                # ressources
                coll=coll,
                doptics=doptics,
                key_diag=key_diag,
                key_cam=k0,
                dsamp=dsamp,
                lamb=lamb,
                # inputs sample points
                x0u=x0u,
                x1u=x1u,
                x0f=x0f,
                x1f=x1f,
                x0l=x0l,
                x1l=x1l,
                dx0=dx0,
                dx1=dx1,
                # options
                sh=sh,
                res_RZ=res_RZ,
                res_phi=res_phi,
                res_rock_curve=res_rock_curve,
                n0=n0,
                n1=n1,
                bool_cross=bool_cross,
                convexHull=convexHull,
                # parameters
                nmax_rays=nmax_rays,
                margin_poly=margin_poly,
                config=config,
                visibility=visibility,
                verb=verb,
                # debug
                debug=debug,
            )

            # -----------
            # add binning

            if dobin is True:
                _dobin(
                    coll=coll,
                    # camera
                    k0=k0,
                    # bins
                    bin0=bin0,
                    bin1=bin1,
                    # lamb
                    lamb=lamb,
                    # dout
                    dout=dout,
                    # remove raw data (lightweight)
                    remove_raw=remove_raw,
                )

            # -------------
            # add input info

            dout[k0]['keym'] = key_mesh
            dout[k0]['res_RZ'] = res_RZ
            dout[k0]['res_phi'] = res_phi
            dout[k0]['res_rock_curve'] = res_rock_curve

    # -------------
    # replace

    if plot is True:
        _plot(
            coll=coll,
            key_diag=key_diag,
            # output
            dout=dout,
            dobin=dobin,
            # plot
            pix0=pix0,
            pix1=pix1,
            tit=tit,
        )

    return dout


# ###########################################################
# ###########################################################
#               check
# ###########################################################


def _check(
    coll=None,
    key_diag=None,
    key_cam=None,
    key_mesh=None,
    # lamb
    lamb=None,
    # config
    config=None,
    # parameters
    res_RZ=None,
    res_phi=None,
    res_rock_curve=None,
    n0=None,
    n1=None,
    convexHull=None,
    # margins
    margin_poly=None,
    # options
    nmax_rays=None,
    # spectro-only
    rocking_curve_fw=None,
    rocking_curve_max=None,
    ## binning
    dobin=None,
    bin0=None,
    bin1=None,
    remove_raw=None,
    # bool
    visibility=None,
    convex=None,
    verb=None,
    debug=None,
    # plot
    plot=None,
    pix0=None,
    pix1=None,
    tit=None,
):

    # --------------------
    # standard vos check
    # --------------------

    (
        key_diag,
        key_mesh,
        spectro,
        is2d,
        doptics,
        dcompute,
        res_RZ,
        res_phi,
        _,       # res_lamb,
        _,       # keep3d,
        _,       # return_vector,
        convexHull,
        visibility,
        verb,
        debug,
        _,       # store,
        _,       # overwrite,
        _,       # timing,
    ) = _vos._check(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        key_mesh=key_mesh,
        res_RZ=res_RZ,
        res_phi=res_phi,
        convexHull=convexHull,
        visibility=visibility,
        verb=verb,
        debug=debug,
    )

    if spectro is False:
        msg = "Routine can only be used for spectro diag, not '{key_diag}'"
        raise Exception(msg)

    # --------------------
    # specific checks
    # --------------------

    # dobin
    dobin = ds._generic_check._check_var(
        dobin, 'dobin',
        default=False,
        types=bool,
    )

    # bin0, bin1
    if dobin is True:

        lc = [
            bin0 is None and bin1 is None,
            np.isscalar(bin0) and np.isscalar(bin1),
            hasattr(bin0, '__iter__') and hasattr(bin1, '__iter__')
        ]
        if np.sum(lc) != 1:
            msg = (
                "If dobin = True, then bin0 and bin1 must either be:\n"
                f"\t- both None:  default to pixels edges\n"
                f"\t- both ints: nb of edges in camera surface\n"
                f"\t- both unique 1d np.ndarray: bins edges\n"
                "Provided:\n"
                f"\t- bin0 = {bin0}\n"
                f"\t- bin1 = {bin1}\n"
            )
            raise Exception(msg)

        if lc[1]:
            bin0, bin1 = int(bin0), int(bin1)
        elif lc[2]:
            bin0 = ds._generic_check._check_flat1darray(
                bin0, 'bin0',
                dtype=float,
                unique=True,
            )
            bin1 = ds._generic_check._check_flat1darray(
                bin1, 'bin1',
                dtype=float,
                unique=True,
            )

    else:
        bin0, nin1 = None, None

    # remove_raw
    remove_raw = ds._generic_check._check_var(
        remove_raw, 'remove_raw',
        default=False,
        types=bool,
    )

    # ------------------
    # check lamb

    # lamb
    lamb = np.unique(ds._generic_check._check_flat1darray(
        lamb, 'lamb',
        dtype=float,
        sign='>0',
    ))

    # --------------------
    # other parameters

    # check nmax_rays
    nmax_rays = int(ds._generic_check._check_var(
        nmax_rays, 'nmax_rays',
        types=(int, float),
        sign='>0',
        default=100000,
    ))

    # plot
    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # pix0, pix1
    if pix0 is not None or pix1 is not None:
        pix0 = ds._generic_check._check_flat1darray(
            pix0, 'pix0',
            dtype=int,
            unique=False,
        )
        pix1 = ds._generic_check._check_flat1darray(
            pix1, 'pix1',
            dtype=int,
            unique=False,
            size=pix0.size,
        )

    # ----------
    # verb

    if verb is True:
        msg = f"\nComputing vos with no bin at lamb for diag '{key_diag}':"
        print(msg)

    return (
        # standard vos
        key_diag,
        key_mesh,
        spectro,
        is2d,
        doptics,
        dcompute,
        res_RZ,
        res_phi,
        convexHull,
        visibility,
        verb,
        debug,
        # specific
        dobin,
        bin0,
        bin1,
        remove_raw,
        lamb,
        nmax_rays,
        plot,
        pix0,
        pix1,
    )


# ###########################################################
# ###########################################################
#               Main
# ###########################################################


def _vos_nobin_at_lamb(
    # ressources
    coll=None,
    doptics=None,
    key_diag=None,
    key_cam=None,
    dsamp=None,
    lamb=None,
    # inputs
    x0u=None,
    x1u=None,
    x0f=None,
    x1f=None,
    x0l=None,
    x1l=None,
    # overall polygons
    pcross0=None,
    pcross1=None,
    phor0=None,
    phor1=None,
    dphi_r=None,
    sh=None,
    res_phi=None,
    res_rock_curve=None,
    n0=None,
    n1=None,
    nmax_rays=None,
    convexHull=None,
    bool_cross=None,
    # parameters
    min_threshold=None,
    margin_poly=None,
    visibility=None,
    verb=None,
    # debug
    debug=None,
    # unused
    **kwdargs,
):
    """ vos computation for spectrometers """

    # -----------------
    # prepare optics

    (
        kspectro,
        lpoly_post,
        p0x, p0y, p0z,
        nin, e0, e1,
        ptsvect_plane,
        ptsvect_spectro,
        ptsvect_cam,
        coords_x01toxyz_plane,
        cent_spectro,
        cent_cam,
        dist_to_cam,
        pix_size,
        cbin0, cbin1,
    ) = _reverse_rt._prepare_optics(
        coll=coll,
        key=key_diag,
        key_cam=key_cam,
    )

    # --------------------------
    # get overall polygons

    pcross0, pcross1 = _utilities._get_overall_polygons(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key_diag]['doptics'],
        key_cam=key_cam,
        poly='pcross',
        convexHull=convexHull,
    )

    phor0, phor1 = _utilities._get_overall_polygons(
        coll=coll,
        doptics=coll.dobj['diagnostic'][key_diag]['doptics'],
        key_cam=key_cam,
        poly='phor',
        convexHull=convexHull,
    )

    # --------------------------
    # add margins

    pcross0, pcross1 = _utilities._get_poly_margin(
        # polygon
        p0=pcross0,
        p1=pcross1,
        # margin
        margin=margin_poly,
    )

    phor0, phor1 = _utilities._get_poly_margin(
        # polygon
        p0=phor0,
        p1=phor1,
        # margin
        margin=margin_poly,
    )

    # ------------------------
    # get ind in cross-section

    pcross = Path(np.array([pcross0, pcross1]).T)
    ind = (
        dsamp['ind']['data']
        & pcross.contains_points(np.array([x0f, x1f]).T).reshape(sh)
    )
    nRZ = ind.sum()

    # R and Z indices
    ir, iz = ind.nonzero()
    iru = np.unique(ir)

    # ----------
    # get dphi_r

    # dphi = doptics[key_cam]['dvos']['dphi']
    # phimin = np.nanmin(dphi[0, :])
    # phimax = np.nanmax(dphi[1, :])

    phi_hor = np.arctan2(phor1, phor0)
    phimin = np.nanmin(phi_hor)
    phimax = np.nanmax(phi_hor)

    # get dphi vs phor
    dphi_r = _utilities._get_dphi_from_R_phor(
        R=x0u[iru],
        phor0=phor0,
        phor1=phor1,
        phimin=phimin,
        phimax=phimax,
        res=res_phi,
        out=True,
    )

    # -------------------------------------
    # prepare lambda, angles, rocking_curve

    (
        nlamb,
        lamb,
        _,
        pow_ratio,
        ang_rel,
        dang,
        angbragg,
    ) = _reverse_rt._prepare_lamb(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        kspectro=kspectro,
        res_lamb=None,
        lamb=lamb,
        res_rock_curve=res_rock_curve,
        verb=verb,
    )

    angbragg0 = angbragg[:1, :]
    angbragg1 = angbragg[-1:, :]

    # linterpbragg = [
        # scpinterp.interp1d(angbragg[:, kk], pow_ratio, kind='linear')
        # for kk in range(angbragg.shape[1])
    # ]

    # --------------
    # prepare output

    shape_x01 = (nmax_rays, nlamb)
    x0 = np.full(shape_x01, np.nan)
    x1 = np.full(shape_x01, np.nan)
    ph_count = np.full(shape_x01, np.nan)

    nmax_real = np.zeros((nlamb,), dtype=int)
    indx01 = np.zeros((nlamb,), dtype=int)

    # ----------
    # verb

    if verb is True:
        msg = (
            f"\tlamb.shape: {lamb.shape}\n"
            f"\tang_rel.shape: {ang_rel.shape}\n"
            f"\tiru.size: {iru.size}\n"
            f"\tnRZ: {nRZ}\n"
        )
        print(msg)

    # ---------------------
    # loop in plasma points

    dr = np.mean(np.diff(x0u))
    dz = np.mean(np.diff(x1u))
    pti = np.r_[0., 0., 0.]
    nru = iru.size

    for i00, i0 in enumerate(iru):

        indiz = ir == i0
        nz = indiz.sum()
        if np.all(np.isnan(dphi_r[:, i00])):
            continue

        nphi = np.ceil(x0u[i0]*(dphi_r[1, i00] - dphi_r[0, i00]) / res_phi).astype(int)
        phir = np.linspace(dphi_r[0, i00], dphi_r[1, i00], nphi)
        cosphi = np.cos(phir)
        sinphi = np.sin(phir)

        dphi = phir[1] - phir[0]
        dv = dr * x0u[i0] * dphi * dz

        for i11, i1 in enumerate(iz[indiz]):

            pti[2] = x1u[i1]

            for i2, phii in enumerate(phir):

                # set point
                pti[0] = x0u[i0]*cosphi[i2]
                pti[1] = x0u[i0]*sinphi[i2]

                # ------------------------------------------
                # initial polygon (crystal on its own plane)

                p0, p1 = ptsvect_plane(
                    pts_x=pti[0],
                    pts_y=pti[1],
                    pts_z=pti[2],
                    vect_x=p0x - pti[0],
                    vect_y=p0y - pti[1],
                    vect_z=p0z - pti[2],
                    strict=True,
                    return_x01=True,
                )[-2:]
                p_a = plg.Polygon(np.array([p0, p1]).T)

                if len(lpoly_post) > 0:
                    # get equivalent aperture
                    p0, p1 = _equivalent_apertures._get_equivalent_aperture(
                        p_a=p_a,
                        pt=pti,
                        nop_pre=len(lpoly_post),
                        lpoly_pre=lpoly_post,
                        ptsvect=ptsvect_plane,
                        min_threshold=min_threshold,
                    )

                    # skip if no intersection
                    if p0 is None or p0.size == 0:
                        continue

                # compute image
                (
                    x0c, x1c, angles, dsang, cosi, iok,
                    dangmin_str, x0if, x1if,
                ) = _reverse_rt._get_points_on_camera_from_pts(
                    p0=p0,
                    p1=p1,
                    pti=pti,
                    # ref
                    cent=cent_spectro,
                    nin=nin,
                    e0=e0,
                    e1=e1,
                    # dang
                    pix_size=pix_size,
                    dist_to_cam=dist_to_cam,
                    dang=dang,
                    phi=phii,
                    # resoluions
                    n0=n0,
                    n1=n1,
                    # functions
                    coords_x01toxyz_plane=coords_x01toxyz_plane,
                    ptsvect_spectro=ptsvect_spectro,
                    ptsvect_cam=ptsvect_cam,
                )[:9]

                if verb is True:
                    msg = (
                        f"\t\t{i00} / {nru}, {i11} / {nz}, {i2} / {nphi}"
                        f":  {iok.sum()} pts   "
                        f"\t dangmin: {dangmin_str}"
                    )
                    print(msg, end='\r')

                if not np.any(iok):
                    continue

                x0c = x0c[iok]
                x1c = x1c[iok]
                angles = angles[iok]
                dsang = dsang[iok]

                # ilamb
                ilamb = (
                    (angles[:, None] >= angbragg0)
                    & (angles[:, None] < angbragg1)
                )

                if not np.any(ilamb):
                    continue

                ilamb_n = np.any(ilamb, axis=0).nonzero()[0]

                for kk in ilamb_n:

                    # npts and indices of pts
                    npts = ilamb[:, kk].sum()

                    # update max
                    nmax_real[kk] += npts

                    # if too many points already => skip
                    if indx01[kk] >= nmax_rays:
                        continue

                    # indices of pts
                    imax = min(indx01[kk] + npts, nmax_rays)
                    indpts = np.arange(indx01[kk], imax)

                    delta = indx01[kk] + npts - nmax_rays
                    if delta > 0:
                        ilamb[ilamb[:, kk].nonzero()[0][-delta:], kk] = False

                    # x0, x1
                    x0[indpts, kk] = x0c[ilamb[:, kk]]
                    x1[indpts, kk] = x1c[ilamb[:, kk]]

                    # indices of power_ratio
                    inds = np.searchsorted(
                        angbragg[:, kk],
                        angles[ilamb[:, kk]],
                    )

                    # update power_ratio * solid angle
                    ph_count[indpts, kk] = (
                        pow_ratio[inds] * dsang[ilamb[:, kk]] * dv
                    )

                    # update index
                    indx01[kk] += npts

    # check nmax_real vs nmax_rays
    imax = nmax_real > nmax_rays
    if np.any(imax):
        lstr = [
            f"\t- {ll}: {nmax_real[ii]} vs {nmax_rays}"
            for ii, ll in enumerate(lamb)
        ]
        msg = (
            "\nSome wavelengths would see more rays than allowed:\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    # remove useless points
    iin = np.any(np.isfinite(ph_count), axis=1)
    if not np.all(iin):
        ph_count = ph_count[iin, :]
        x0 = x0[iin, :]
        x1 = x1[iin, :]

    # remove useless lamb
    iin = ph_count > 0.
    ilamb = np.any(np.any(np.any(iin, axis=0), axis=0), axis=0)
    if not np.all(ilamb):
        ph_count = ph_count[..., ilamb]
        lamb = lamb[ilamb]

    # --------------------
    # return

    # dout
    dout = {
        # lamb
        'lamb': {
            'data': lamb,
            'units': 'm',
        },
        # data
        'x0': {'data': x0, 'units': 'm'},
        'x1': {'data': x1, 'units': 'm'},
        'ph_count': {
            'data': ph_count,
            'units': 'sr.m3',
        },
        # information
        'nmax_real': {'data': nmax_real, 'units': 'nb. rays'},
        'nmax_rays': {'data': nmax_rays, 'units': 'nb. rays'},
    }

    return dout


# ################################################
# ################################################
#           do bin
# ################################################


def _dobin(
    coll=None,
    # camera
    k0=None,
    # bins
    bin0=None,
    bin1=None,
    # lamb
    lamb=None,
    # dout
    dout=None,
    # remove raw data
    remove_raw=None,
):

    # ----------------------------------
    # get bins (default = camera pixels)

    c0, c1 = coll.dobj['camera'][k0]['dgeom']['cents']
    if isinstance(bin0, np.ndarray) and isinstance(bin1, np.ndarray):
        pass

    else:
        c0 = coll.ddata[c0]['data']
        c1 = coll.ddata[c1]['data']

        c0min = c0[0] - 0.5*(c0[1] - c0[0])
        c0max = c0[-1] + 0.5*(c0[-1] - c0[-2])
        c1min = c1[0] - 0.5*(c1[1] - c1[0])
        c1max = c1[-1] + 0.5*(c1[-1] - c1[-2])

        if bin0 is None or bin1 is None:
            bin0 = np.r_[c0min, 0.5*(c0[1:] + c0[:-1]), c0max]
            bin1 = np.r_[c1min, 0.5*(c1[1:] + c1[:-1]), c1max]

        elif isinstance(bin0, int) and isinstance(bin1, int):
            bin0 = np.linspace(c0min, c0max, bin0)
            bin1 = np.linspace(c1min, c1max, bin1)

        else:
            raise Exception("Impossible")

    # --------------------------
    # derive bin centers in 2d

    c0 = 0.5 * (bin0[1:] + bin0[:-1])
    c1 = 0.5 * (bin1[1:] + bin1[:-1])

    n0, n1 = c0.size, c1.size

    # -----------------------
    # compute

    # points with weight
    bin_ph = np.full((n0, n1, lamb.size), 0.)
    for kk, ll in enumerate(lamb):

        iok = np.isfinite(dout[k0]['x0']['data'][:, kk])
        if not np.any(iok):
            continue

        bin_ph[..., kk] = scpstats.binned_statistic_2d(
            dout[k0]['x0']['data'][iok, kk],
            dout[k0]['x1']['data'][iok, kk],
            dout[k0]['ph_count']['data'][iok, kk],
            statistic='sum',
            bins=(bin0, bin1),
        ).statistic

    # --------
    # store

    dout[k0]['bin0'] = {'data': bin0, 'units': dout[k0]['x0']['units']}
    dout[k0]['bin1'] = {'data': bin1, 'units': dout[k0]['x1']['units']}
    dout[k0]['bin_ph'] = {'data': bin_ph, 'units': dout[k0]['ph_count']['units']}

    if remove_raw is True:
        for k1 in ['x0', 'x1', 'ph_count']:
            del dout[k0][k1]

    return


# ################################################
# ################################################
#           plot
# ################################################


def _plot(
    coll=None,
    key_diag=None,
    # output
    dout=None,
    dobin=None,
    # plot
    pix0=None,
    pix1=None,
    tit=None,
):

    # -------------
    # prepare

    if tit is None:
        tit = (
            f"vos ray tracing for '{key_diag}'"
        )

    # ---------------------
    # prepare figure / axes

    dmargins = {
        'left': 0.08, 'right': 0.95,
        'bottom': 0.06, 'top': 0.90,
        'wspace': 0.5, 'hspace': 0.2,
    }

    nax = 20
    fig = plt.figure(figsize=(14, 8))
    fig.suptitle(tit, size=12, fontweight='bold')
    gs = gridspec.GridSpec(ncols=nax*len(dout), nrows=2, **dmargins)

    dax = {}
    for ii, (k0, v0) in enumerate(dout.items()):

        # -------------------
        # prepare image

        if dobin is True:
            extent = (
                v0['bin0']['data'][0], v0['bin0']['data'][-1],
                v0['bin1']['data'][0], v0['bin1']['data'][-1],
            )

            im = np.sum(v0['bin_ph']['data'], axis=-1)
            im[im == 0.] = np.nan

        # ------------
        # prepare axes

        ax0 = fig.add_subplot(gs[0, ii*nax:(ii+1)*nax-2], aspect='equal', adjustable='datalim')
        ax0.set_xlabel('x0 (m)', size=12, fontweight='bold')
        ax0.set_ylabel('x1 (m)', size=12, fontweight='bold')

        ax1 = fig.add_subplot(gs[1, ii*nax:(ii+1)*nax-2], aspect='equal', adjustable='datalim')
        ax1.set_xlabel('x0 (m)', size=12, fontweight='bold')
        ax1.set_ylabel('x1 (m)', size=12, fontweight='bold')

        cax = fig.add_subplot(gs[1, (ii+1)*nax-2])

        dax[f'ax0_{k0}'] = {'handle': ax0}
        dax[f'ax1_{k0}'] = {'handle': ax1}

        # -----------------
        # plot

        # get total camera outline
        out0, out1 = coll.get_optics_outline(k0, total=True, closed=True)

        # camera outline
        ax0.plot(out0, out1, '-k')
        ax1.plot(out0, out1, '-k')

        # points without weight
        for kk, ll in enumerate(v0['lamb']['data']):
            ax0.plot(
                v0['x0']['data'][:, kk],
                v0['x1']['data'][:, kk],
                ls='None',
                marker='.',
                label=r"$\lambda = $" + f" {ll*1e10:5.3f} A",
            )

        # points with weight
        if dobin is True:
            im0 = ax1.imshow(
                im.T,
                extent=extent,
                aspect='equal',
                origin='lower',
                interpolation='nearest',
                vmin=0,
                vmax=np.nanmax(im),
                cmap=plt.cm.viridis,
            )

            cb = plt.colorbar(
                im0,
                ax=ax1,
                cax=cax,
            )
            cb.set_label(label=v0['bin_ph']['units'], size=12, weight='bold')


        # pixel
        if pix0 is not None:

            # get pixel outline
            out0, out1 = coll.get_optics_outline(k0, total=False, closed=True)

            # centers
            c0, c1 = coll.dobj['camera'][k0]['dgeom']['cents']
            if dobin is False:
                c0 = coll.ddata[c0]['data'][pix0]
                c1 = coll.ddata[c1]['data'][pix1]
            else:
                # adjust pixel outline
                out0 *= coll.ddata[c0]['data'].size / (dout[k0]['bin0']['data'].size - 1)
                out1 *= coll.ddata[c1]['data'].size / (dout[k0]['bin1']['data'].size - 1)

                # get pixel centers
                c0 = 0.5*(dout[k0]['bin0']['data'][1:] + dout[k0]['bin0']['data'][:-1])[pix0]
                c1 = 0.5*(dout[k0]['bin1']['data'][1:] + dout[k0]['bin1']['data'][:-1])[pix1]

            # multiple pixels
            out0 = (c0[:, None] + np.r_[out0, np.nan][None, :]).ravel()
            out1 = (c1[:, None] + np.r_[out1, np.nan][None, :]).ravel()

            ax0.plot(out0, out1, '-k')
            ax1.plot(out0, out1, '-k')

        ax0.legend()

    return dax


"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import shutil
import itertools as itt
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu as tf


_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.mesh.test_01_checks'
_TOFU_USER = os.path.join(os.path.expanduser("~"), '.tofu')
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(_here)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)
    # Recreating clean .tofu
    # out = subprocess.run(_CUSTOM, stdout=PIPE, stderr=PIPE)
    os.system('python '+_CUSTOM)


def teardown_module():
    print("Removing user ~/.tofu/ if any")
    if os.path.isdir(_TOFU_USER):
        shutil.rmtree(_TOFU_USER)


#######################################################
#
#     checking routines
#
#######################################################


class Test01_Inversions():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):

        # create conf and cam
        conf0 = tf.load_config('WEST-V0')

        coll = tf.data.Collection()

        # add camera
        npix = 10
        coll.add_camera_pinhole(
            key='camH',
            key_diag='d0',
            x=3.0,
            y=1.,
            z=0.3,
            pinhole_size=0.01,
            focal=0.1,
            pix_size=0.1,
            pix_nb=npix,
            theta=-5*np.pi/6,
            dphi=0,
            tilt=0,
            config=conf0,
            compute=False,
        )

        coll.add_camera_pinhole(
            key='camV',
            key_diag='d0',
            x=3.0,
            y=1.,
            z=-0.3,
            pinhole_size=0.01,
            focal=0.1,
            pix_size=0.1,
            pix_nb=npix,
            theta=5*np.pi/6,
            dphi=0,
            tilt=0,
            config=conf0,
        )

        coll.add_camera_pinhole(
            key='cam2',
            cam_type='2d',
            key_diag='d1',
            x=3.0,
            y=1.,
            z=-0.3,
            pinhole_size=0.01,
            focal=0.1,
            pix_size=0.1,
            pix_nb=[10, 5],
            theta=5*np.pi/6,
            dphi=0,
            tilt=0,
            config=conf0,
        )

        # mesh rect deg 1 and 2
        coll.add_mesh_2d_rect(
            crop_poly=conf0,
            key='m1',
            res=0.10,
            deg=0,
        )
        coll.add_bsplines(deg=1)
        coll.add_bsplines(deg=2)

        # add 2d radius for polar mesh
        kR, kZ = coll.dobj['mesh']['m1']['knots']
        R, Z = coll.ddata[kR]['data'], coll.ddata[kZ]['data']
        nR, nZ = R.size, Z.size
        R = np.repeat(R[:, None], nZ, axis=1)
        Z = np.repeat(Z[None, :], nR, axis=0)
        rad2d = ((R-2.4)/0.4)**2 + (Z/0.5)**2
        krad = 'rad2d'
        coll.add_data(key=krad, data=rad2d, ref='m1_bs1')

        # mesh polar deg 1 and 2
        coll.add_mesh_1d(
            key='m2',
            knots=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 1.],
            subkey=krad,
            deg=0,
        )
        coll.add_bsplines(key='m2', deg=1)
        coll.add_bsplines(key='m2', deg=2)

        # add geometry matrices
        # coll.add_ref(key='chan', data=chan, group='chan')
        coll.add_geometry_matrix(key_diag='d0', key_bsplines='m1_bs0')
        coll.add_geometry_matrix(key_diag='d0', key_bsplines='m1_bs1')
        coll.add_geometry_matrix(key_diag='d1', key_bsplines='m1_bs2')
        coll.add_geometry_matrix(key_diag='d0', key_bsplines='m2_bs0')
        coll.add_geometry_matrix(key_diag='d0', key_bsplines='m2_bs1')
        coll.add_geometry_matrix(key_diag='d1', key_bsplines='m2_bs2')

        # add emiss
        t0 = np.array([0, 1])
        kap = coll.dobj['bsplines']['m2_bs1']['apex'][0]
        rad = coll.ddata[kap]['data']
        emiss = (
            np.exp(-(rad)**2/0.2**2)
            + 0.1*np.cos(t0)[:, None]*np.exp(-rad**2/0.05**2)
        )

        coll.add_ref(key='nt0', size=2)
        coll.add_data(key='t0', data=t0, dim='time', ref='nt0', units='s')
        coll.add_data(
            key='emiss',
            data=emiss,
            ref=('nt0', 'm2_bs1'),
            units='W/(m3.sr)',
        )

        # add synthetic data
        coll.compute_diagnostic_signal(
            key='s0',
            key_diag='d0',
            key_integrand='emiss',
            res=0.01,
        )

        coll.compute_diagnostic_signal(
            key='s1',
            key_diag='d1',
            key_integrand='emiss',
            res=0.01,
        )

        self.coll = coll

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_run_all_and_plot(self):

        dalgo = tf.data.get_available_inversions_algo(returnas=dict)
        lstore = [True, False]

        # running
        lkmat = list(self.coll.dobj['geom matrix'].keys())
        for ii, kmat in enumerate(lkmat):

            kbs = self.coll.dobj['geom matrix'][kmat]['bsplines']
            kd = self.coll.dobj['geom matrix'][kmat]['diagnostic']
            km = self.coll.dobj['bsplines'][kbs]['mesh']
            nd = self.coll.dobj['mesh'][km]['nd']
            mtype = self.coll.dobj['mesh'][km]['type']
            submesh = self.coll.dobj['mesh'][km]['submesh']
            deg = self.coll.dobj['bsplines'][kbs]['deg']

            if deg in [0, 1]:
                lop = ['D1N2']
            else:
                lop = ['D1N2', 'D2N2']

            for jj, comb in enumerate(itt.product(dalgo.keys(), lop)):

                if comb[0] == 'algo5':
                    continue

                if comb[1] == 'D2N2' and deg != 2:
                    continue

                if 'mfr' in comb[0].lower() and deg != 0:
                    continue

                algofam = dalgo[comb[0]]['family']
                if algofam == 'Non-regularized' and nd != '1d':
                    continue

                if kmat == 'gmat05' and comb == ('algo1', 'D2N2'):
                    continue

                if comb == ('algo1', 'D1N2'):
                    continue

                dconstraints = None
                if km == 'm2':
                    if jj == 1:
                        dconstraints = {'deriv1': {'rad': 0, 'val': 0}}
                    elif jj == 2:
                        dconstraints = {'rmax': 0.70}

                kdat = 's0' if kd == 'd0' else 's1'
                self.coll.add_inversion(
                    algo=comb[0],
                    key_matrix=kmat,
                    key_data=kdat,
                    sigma=0.10,
                    operator=comb[1],
                    store=jj%2 == 0,
                    conv_crit=1.e-3,
                    kwdargs={'tol': 1.e-2, 'maxiter': 100},
                    maxiter_outer=10,
                    dref_vector={'units': 's'},
                    dconstraints=dconstraints,
                    verb=1,
                )
                ksig = f'{kdat}-sigma'
                if ksig in self.coll.ddata.keys():
                    self.coll.remove_data(ksig)

        # plotting
        linv = list(self.coll.dobj['inversions'].keys())[::7]
        for kinv in linv:
            dax = self.coll.plot_inversion(
                key=kinv,
                res=0.1,
                dref_vector={'units': 's'},
            )
            plt.close('all')



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import os


import numpy as np
import matplotlib.pyplot as plt


import tofu as tf


_PATH_HERE = os.path.dirname(__file__)
_PATH_TESTDATA = os.path.join(
    os.path.dirname(_PATH_HERE),
    'tests',
    'tests06_mesh',
    'test_data',
)
_PFE_TESTDATA = os.path.join(_PATH_TESTDATA, 'mesh_triangular_WEST_eq.txt')


# #############################################################################
# #############################################################################
#               Benchmark of Mesh2D routines
#                   High level routines
# #############################################################################


class HighLevel:
    """ Benchmark the geometry-oriented routines

    In particular:
        - camera creation (ray-tracing)
        - solid angle computing

    """

    # -----------------------------
    # Attributes reckognized by asv

    # time before benchmark is killed
    timeout = 30
    repeat = (1, 10, 20.0)
    sample_time = 0.100

    # -------------------------------------------------------
    # Setup and teardown, run before / after benchmark methods

    def setup_cache(self):

        # prepare input dict for a cam2d
        dcam2d = {
            'pinhole': [8.38, 0., 0.],
            'orientation': [-7*np.pi/8, np.pi/6, 0],
            'focal': 0.08,
            'sensor_nb': 200,
            'sensor_size': 0.2,
            'Diag': 'SXR',
            'Exp': 'WEST',
            'Name': 'cam2',
        }

        # prepare input dict for mesh2d rect
        dmesh2drect = {
            'res': 0.10,
            'deg': 2,
            'key': 'rect1',
        }

        # prepare input dict for mesh2d tri
        out = np.loadtxt(_PFE_TESTDATA)
        dmesh2dtri = {
            'knots': out[1:int(out[0, 0])+1, :2],
            'cents': out[int(out[0, 0])+1:, :].astype(int),
            'deg': 1,
            'key': 'tri1',
        }

        return dcam2d, dmesh2drect, dmesh2dtri

    def setup(self, out):
        """ run before each benchmark method, out from setup_cache  """
        self.conf = tf.load_config('ITER')
        self.mesh2d = tf.data.Mesh2D()
        self.mesh2d.add_mesh(crop_poly=self.conf, **out[1])
        self.mesh2d.add_mesh(**out[2])
        self.cam = tf.geom.utils.create_CamLOS1D(config=self.conf, **out[0])

    def teardown(self, out):
        pass

    # -------------------------------------
    # benchmarks methods for geometry tools

    # Mesh2D - rect
    def time_00_mesh2d_rect_bs2(self, out):
        mesh = tf.data.Mesh2D()
        mesh.add_mesh(crop_poly=self.conf, **out[1])

    # Mesh2D - tri
    def time_01_mesh2dtri_bs1(self, out):
        mesh = tf.data.Mesh2D()
        mesh.add_mesh(**out[2])

    # Geometry matrix - rect
    def time_02_geommatrix_rect(self, out):
        self.mesh2d.add_geometry_matrix(
            key='rect1-bs2', cam=self.cam, res=0.01, verb=False,
        )



# -*- coding: utf-8 -*-
"""
Provide spectrum fitting functions (storing, processing, plotting...)
"""
from ._fit12d import *
from ._analysis_tools import *
from ._plot import *
from ._rockingcurve import *
from ._rockingcurve_tools import *
from ._spectralrange2d import *
